"""SC Controller - Global Settings.

Currently setups only one thing...
"""

import json
import logging
import os
import re
import sys
import traceback

from gi.repository import Gdk, GdkPixbuf, GLib, GObject, Gtk

from scc.actions import Action, NoAction
from scc.constants import LEFT, RIGHT
from scc.gui.dwsnc import IS_UNITY
from scc.gui.editor import ComboSetter, Editor
from scc.gui.osk_binding_editor import OSKBindingEditor
from scc.gui.parser import GuiActionParser
from scc.gui.userdata_manager import UserDataManager
from scc.menu_data import MenuData, MenuGenerator, MenuItem, Separator, Submenu
from scc.modifiers import SensitivityModifier
from scc.osd.keyboard import Keyboard as OSDKeyboard
from scc.osd.menu_generators import RecentListMenuGenerator
from scc.osd.osk_actions import OSKCursorAction
from scc.paths import get_config_path, get_menus_path, get_profiles_path, get_share_path
from scc.profile import Encoder, Profile
from scc.special_actions import ChangeProfileAction, RestartDaemonAction, TurnOffAction
from scc.tools import _, find_binary, find_menu, find_profile
from scc.x11.autoswitcher import AutoSwitcher, Condition

log = logging.getLogger("GS")


class GlobalSettings(Editor, UserDataManager, ComboSetter):
	GLADE = "global_settings.glade"

	DEFAULT_MENU_OPTIONS = [
		# label,				order, class, icon, parameter
		("Recent profiles", 0, RecentListMenuGenerator, None, 3),
		("Autoswitch Options", 1, Submenu, "system/autoswitch", ".autoswitch.menu"),
		("Switch To", 1, Submenu, "system/windowlist", ".windowlist.menu"),
		("Display Keyboard", 2, MenuItem, "system/keyboard", "keyboard()"),
		("Turn Controller OFF", 2, MenuItem, "system/turn-off", "osd(turnoff())"),
		(
			"Kill Current Window",
			1,
			MenuItem,
			"weapons/pistol-gun",
			"dialog('Really? Non-saved progress or data will be lost', "
			"name('Back', None), "
			"name('Kill', shell('kill -9 $(xdotool getwindowfocus getwindowpid)')))",
		),
		("Run Program...", 1, MenuItem, "system/cog", 'shell("scc-osd-launcher")'),
		("Display Current Bindings...", 1, MenuItem, "system/binding-display", 'shell("scc-osd-show-bindings")'),
		("Games", 1, Submenu, "system/controller", ".games.menu"),
		("Edit Bindings", 2, MenuItem, "system/edit", 'shell("sc-controller --osd")'),
		# order: 0 - top, 1 - after 'options', 2 bottom
	]

	def __init__(self, app):
		UserDataManager.__init__(self)
		self.app = app
		self.setup_widgets()
		self._timer = None
		self._recursing = False
		self._gamepad_icons = {
			"unknown": GdkPixbuf.Pixbuf.new_from_file(
				os.path.join(self.app.imagepath, "controller-icons", "unknown.svg"),
			),
		}
		self.app.config.reload()
		Action.register_all(sys.modules["scc.osd.osk_actions"], prefix="OSK")
		self.load_settings()
		self.load_profile_list()
		self._recursing = False
		self._eh_ids = (self.app.dm.connect("reconfigured", self.on_daemon_reconfigured),)

	def _get_gamepad_icon(self, drv):
		if drv in self._gamepad_icons:
			return self._gamepad_icons[drv]
		try:
			p = GdkPixbuf.Pixbuf.new_from_file(os.path.join(self.app.imagepath, "controller-icons", drv + "-4.svg"))
		except:
			log.warning("Failed to load gamepad icon for driver '%s'", drv)
			p = self._gamepad_icons["unknown"]
		self._gamepad_icons[drv] = p
		return p

	def on_daemon_reconfigured(self, *a):
		# config is reloaded in main window 'reconfigured' handler.
		# Using GLib.idle_add here ensures that main window hanlder will run
		# *before* self.load_conditions
		GLib.idle_add(self.load_settings)

	def on_Dialog_destroy(self, *a):
		for x in self._eh_ids:
			self.app.dm.disconnect(x)
		self._eh_ids = ()
		Action.unregister_prefix("OSK")

	def load_settings(self):
		self.load_autoswitch()
		self.load_osk()
		self.load_colors()
		self.load_cbMIs()
		self.load_drivers()
		self.load_controllers()
		# Load rest
		self._recursing = True
		(self.builder.get_object("cbInputTestMode").set_active(bool(self.app.config["enable_sniffing"])))
		(self.builder.get_object("cbEnableSerials").set_active(not bool(self.app.config["ignore_serials"])))
		(self.builder.get_object("cbEnableRumble").set_active(bool(self.app.config["output"]["rumble"])))
		(self.builder.get_object("cbEnableStatusIcon").set_active(bool(self.app.config["gui"]["enable_status_icon"])))
		(
			self.builder.get_object("cbMinimizeToStatusIcon").set_active(
				not IS_UNITY and bool(self.app.config["gui"]["minimize_to_status_icon"]),
			)
		)
		(
			self.builder.get_object("cbMinimizeToStatusIcon").set_sensitive(
				not IS_UNITY and self.app.config["gui"]["enable_status_icon"],
			)
		)
		(
			self.builder.get_object("cbMinimizeOnStart").set_active(
				not IS_UNITY and bool(self.app.config["gui"]["minimize_on_start"]),
			)
		)
		(
			self.builder.get_object("cbMinimizeOnStart").set_sensitive(
				not IS_UNITY and self.app.config["gui"]["enable_status_icon"],
			)
		)
		(self.builder.get_object("cbAutokillDaemon").set_active(self.app.config["gui"]["autokill_daemon"]))
		(self.builder.get_object("cbNewRelease").set_active(self.app.config["gui"]["news"]["enabled"]))
		self._recursing = False

		try:
			import evdev
		except ImportError:
			# This block runs if evdev module is missing
			for w in ("cbEnableDriver_evdevdrv", "btAddController"):
				self.builder.get_object(w).set_sensitive(False)
			self.builder.get_object("txEvdevMissing").set_visible(True)

	def load_drivers(self):
		for key, value in self.app.config["drivers"].items():
			w = self.builder.get_object("cbEnableDriver_%s" % (key,))
			if w:
				w.set_active(value)

	def _load_color(self, w, dct, key):
		"""Common part of load_colors"""
		if w:
			success, color = Gdk.Color.parse("#%s" % (self.app.config[dct][key],))
			if not success:
				success, color = Gdk.Color.parse("#%s" % (self.app.config[dct][key],))
			w.set_color(color)

	def load_colors(self):
		cbOSDStyle = self.builder.get_object("cbOSDStyle")
		cbOSDColorPreset = self.builder.get_object("cbOSDColorPreset")
		for k in self.app.config["osd_colors"]:
			w = self.builder.get_object("cb%s" % (k,))
			self._load_color(w, "osd_colors", k)
		for k in self.app.config["osk_colors"]:
			w = self.builder.get_object("cbosk_%s" % (k,))
			self._load_color(w, "osk_colors", k)
		theme = self.app.config.get("osd_color_theme", "None")
		self.set_cb(cbOSDColorPreset, theme)
		self.set_cb(cbOSDStyle, self.app.config.get("osd_style"))

	def load_autoswitch(self):
		"""Transfers autoswitch settings from config to UI"""
		tvItems = self.builder.get_object("tvItems")
		cbShowOSD = self.builder.get_object("cbShowOSD")
		conditions = AutoSwitcher.parse_conditions(self.app.config)
		model = tvItems.get_model()
		model.clear()
		for cond in conditions.keys():
			o = GObject.GObject()
			o.condition = cond
			o.action = conditions[cond]
			a_str = o.action.describe(Action.AC_SWITCHER)
			model.append((o, o.condition.describe(), a_str))
		self._recursing = True
		self.on_tvItems_cursor_changed()
		cbShowOSD.set_active(bool(self.app.config["autoswitch_osd"]))
		self._recursing = False

	def load_osk(self):
		cbStickAction = self.builder.get_object("cbStickAction")
		cbTriggersAction = self.builder.get_object("cbTriggersAction")
		profile = Profile(GuiActionParser())
		profile.load(find_profile(OSDKeyboard.OSK_PROF_NAME))
		self._recursing = True

		# Load triggers
		triggers = "%s|%s" % (profile.triggers[LEFT].to_string(), profile.triggers[RIGHT].to_string())
		if not self.set_cb(cbTriggersAction, triggers, keyindex=1):
			self.add_custom(cbTriggersAction, triggers)

		# Load stick
		if not self.set_cb(cbStickAction, profile.stick.to_string(), keyindex=1):
			self.add_custom(cbStickAction, profile.stick.to_string())

		# Load sensitivity
		s = profile.pads[LEFT].compress().speed
		self.builder.get_object("sclSensX").set_value(s[0])
		self.builder.get_object("sclSensY").set_value(s[1])

		self._recursing = False

	def add_custom(self, cb, key):
		for k in cb.get_model():
			if k[2]:
				k[1] = key
				self.set_cb(cb, key, keyindex=1)
				return
		cb.get_model().append((_("(customized)"), key, True))
		self.set_cb(cb, key, keyindex=1)

	def _load_osk_profile(self):
		"""Loads and returns on-screen keyboard profile.
		Used by methods that are changing it.
		"""
		profile = Profile(GuiActionParser())
		profile.load(find_profile(OSDKeyboard.OSK_PROF_NAME))
		return profile

	def _save_osk_profile(self, profile):
		"""Saves on-screen keyboard profile and calls daemon.reconfigure()
		Used by methods that are changing it.
		"""
		profile.save(os.path.join(get_profiles_path(), OSDKeyboard.OSK_PROF_NAME + ".sccprofile"))
		self.app.dm.reconfigure()

	def on_cbStickAction_changed(self, cb):
		if self._recursing:
			return
		key = cb.get_model().get_value(cb.get_active_iter(), 1)
		profile = self._load_osk_profile()
		profile.stick = GuiActionParser().restart(key).parse()
		self._save_osk_profile(profile)

	def on_cbTriggersAction_changed(self, cb):
		if self._recursing:
			return
		key = cb.get_model().get_value(cb.get_active_iter(), 1)
		l, r = key.split("|")
		profile = self._load_osk_profile()
		profile.triggers[LEFT] = GuiActionParser().restart(l).parse()
		profile.triggers[RIGHT] = GuiActionParser().restart(r).parse()
		self._save_osk_profile(profile)

	def on_osd_color_set(self, *a):
		"""Called when user selects color.
		"""
		# Following lambdas converts Gdk.Color into #rrggbb notation.
		# Gdk.Color can do similar, except it uses #rrrrggggbbbb notation that
		# is not understood by Gdk css parser....
		cbOSDColorPreset = self.builder.get_object("cbOSDColorPreset")
		striphex = lambda a: hex(a).strip("0x").zfill(2)
		tohex = lambda a: "".join([striphex(int(x * 0xFF)) for x in a.to_floats()])
		for k in self.app.config["osd_colors"]:
			w = self.builder.get_object("cb%s" % (k,))
			if w:
				self.app.config["osd_colors"][k] = tohex(w.get_color())
		for k in self.app.config["osk_colors"]:
			w = self.builder.get_object("cbosk_%s" % (k,))
			if w:
				self.app.config["osk_colors"][k] = tohex(w.get_color())
		self.app.config["osd_color_theme"] = None
		self.set_cb(cbOSDColorPreset, "None")
		self.app.save_config()

	def schedule_save_config(self):
		"""Schedules config saving in 3s.
		Done to prevent literal madness when user moves slider.
		"""

		def cb(*a):
			self._timer = None
			self.app.save_config()

		if self._timer is not None:
			GLib.source_remove(self._timer)
		self._timer = GLib.timeout_add_seconds(3, cb)

	def save_config(self) -> None:
		"""Transfers settings from UI back to config."""
		# Store hard stuff
		tvItems = self.builder.get_object("tvItems")
		cbShowOSD = self.builder.get_object("cbShowOSD")
		cbEnableStatusIcon = self.builder.get_object("cbEnableStatusIcon")
		cbMinimizeToStatusIcon = self.builder.get_object("cbMinimizeToStatusIcon")
		conds = []
		for row in tvItems.get_model():
			conds.append({"condition": row[0].condition.encode(), "action": row[0].action.to_string()})
		# Apply status icon settings
		if self.app.config["gui"]["enable_status_icon"] != cbEnableStatusIcon.get_active():
			self.app.config["gui"]["enable_status_icon"] = cbEnableStatusIcon.get_active()
			cbMinimizeToStatusIcon.set_sensitive(not IS_UNITY and cbEnableStatusIcon.get_active())
			if cbEnableStatusIcon.get_active():
				self.app.setup_statusicon()
			else:
				self.app.destroy_statusicon()
		# Store rest
		self.app.config["autoswitch"] = conds
		self.app.config["autoswitch_osd"] = cbShowOSD.get_active()
		self.app.config["enable_sniffing"] = self.builder.get_object("cbInputTestMode").get_active()
		self.app.config["ignore_serials"] = not self.builder.get_object("cbEnableSerials").get_active()
		self.app.config["output"]["rumble"] = self.builder.get_object("cbEnableRumble").get_active()
		self.app.config["gui"]["enable_status_icon"] = self.builder.get_object("cbEnableStatusIcon").get_active()
		self.app.config["gui"]["minimize_to_status_icon"] = self.builder.get_object(
			"cbMinimizeToStatusIcon",
		).get_active()
		self.app.config["gui"]["minimize_on_start"] = self.builder.get_object("cbMinimizeOnStart").get_active()
		self.app.config["gui"]["autokill_daemon"] = self.builder.get_object("cbAutokillDaemon").get_active()
		self.app.config["gui"]["news"]["enabled"] = self.builder.get_object("cbNewRelease").get_active()

		# Save
		self.app.save_config()

	def on_cbShowOSD_toggled(self, cb):
		if self._recursing:
			return
		self.save_config()

	def on_btRestartEmulation_clicked(self, *a):
		rvRestartWarning = self.builder.get_object("rvRestartWarning")
		self.app.dm.stop()
		rvRestartWarning.set_reveal_child(False)
		GLib.timeout_add_seconds(1, self.app.dm.start)

	def on_restarting_checkbox_toggled(self, *a):
		if self._recursing:
			return
		self.on_random_checkbox_toggled()
		self._needs_restart()

	def _needs_restart(self):
		if self.app.dm.is_alive():
			rvRestartWarning = self.builder.get_object("rvRestartWarning")
			rvRestartWarning.set_reveal_child(True)

	DRIVER_DEPS = {"ds4drv": ("evdevdrv", "hiddrv")}

	def on_cbEnableDriver_toggled(self, cb):
		if self._recursing:
			return
		drv = cb.get_name()
		self.app.config["drivers"][drv] = cb.get_active()
		if cb.get_active() and drv in self.DRIVER_DEPS:
			# Driver has dependencies, make sure at least one of them is active
			one_active = any([self.app.config["drivers"].get(x) for x in self.DRIVER_DEPS[drv]])
			if not one_active:
				# Nothing is, make everything active just to be sure
				self._recursing = True
				for x in self.DRIVER_DEPS[drv]:
					w = self.builder.get_object("cbEnableDriver_%s" % (x,))
					if w:
						w.set_active(True)
					self.app.config["drivers"][x] = True
				self._recursing = False

		if not cb.get_active() and any([drv in x for x in self.DRIVER_DEPS.values()]):
			# Something depends on this driver,
			# disable anything that has no dependent drivers active
			self._recursing = True
			for x, deps in self.DRIVER_DEPS.items():
				w = self.builder.get_object("cbEnableDriver_%s" % (x,))
				one_active = any([self.app.config["drivers"].get(y) for y in self.DRIVER_DEPS[x]])
				if not one_active and w:
					w.set_active(False)
					self.app.config["drivers"][x] = False
			self._recursing = False

		self.save_config()
		self._needs_restart()

	def on_random_checkbox_toggled(self, *a):
		if self._recursing:
			return
		self.save_config()

	def on_butEditKeyboardBindings_clicked(self, *a):
		e = OSKBindingEditor(self.app)
		e.show(self.window)

	def btEdit_clicked_cb(self, *a):
		"""Handler for "Edit condition" button"""
		tvItems = self.builder.get_object("tvItems")
		cbProfile = self.builder.get_object("cbProfile")
		ce = self.builder.get_object("ConditionEditor")
		entTitle = self.builder.get_object("entTitle")
		entClass = self.builder.get_object("entClass")
		cbMatchTitle = self.builder.get_object("cbMatchTitle")
		cbMatchClass = self.builder.get_object("cbMatchClass")
		cbExactTitle = self.builder.get_object("cbExactTitle")
		cbRegExp = self.builder.get_object("cbRegExp")
		rbProfile = self.builder.get_object("rbProfile")
		rbTurnOff = self.builder.get_object("rbTurnOff")
		rbRestart = self.builder.get_object("rbRestart")
		# Grab data
		model, iter = tvItems.get_selection().get_selected()
		o = model.get_value(iter, 0)
		profile = model.get_value(iter, 2)
		condition = o.condition
		action = o.action

		# Clear editor
		for cb in (cbMatchTitle, cbMatchClass, cbExactTitle, cbRegExp):
			cb.set_active(False)
		for ent in (entClass, entTitle):
			ent.set_text("")
		# Setup action
		if isinstance(o.action, ChangeProfileAction):
			rbProfile.set_active(True)
			self.set_cb(cbProfile, o.action.profile)
		elif isinstance(o.action, TurnOffAction):
			rbTurnOff.set_active(True)
		elif isinstance(o.action, RestartDaemonAction):
			rbRestart.set_active(True)

		# Setup editor
		if condition.title:
			entTitle.set_text(condition.title or "")
			cbMatchTitle.set_active(True)
		elif condition.exact_title:
			entTitle.set_text(condition.exact_title or "")
			cbExactTitle.set_active(True)
			cbMatchTitle.set_active(True)
		elif condition.regexp:
			try:
				entTitle.set_text(condition.regexp.pattern)
			except:
				entTitle.set_text("")
			cbRegExp.set_active(True)
			cbMatchTitle.set_active(True)
		if condition.wm_class:
			entClass.set_text(condition.wm_class or "")
			cbMatchClass.set_active(True)

		# Show editor
		ce.show()

	def on_btSave_clicked(self, *a):
		tvItems = self.builder.get_object("tvItems")
		cbProfile = self.builder.get_object("cbProfile")
		entTitle = self.builder.get_object("entTitle")
		entClass = self.builder.get_object("entClass")
		cbMatchTitle = self.builder.get_object("cbMatchTitle")
		cbMatchClass = self.builder.get_object("cbMatchClass")
		cbExactTitle = self.builder.get_object("cbExactTitle")
		cbRegExp = self.builder.get_object("cbRegExp")
		rbProfile = self.builder.get_object("rbProfile")
		rbTurnOff = self.builder.get_object("rbTurnOff")
		rbRestart = self.builder.get_object("rbRestart")
		ce = self.builder.get_object("ConditionEditor")

		# Build condition
		data = {}
		if cbMatchTitle.get_active() and entTitle.get_text():
			if cbExactTitle.get_active():
				data["exact_title"] = entTitle.get_text()
			elif cbRegExp.get_active():
				data["regexp"] = entTitle.get_text()
			else:
				data["title"] = entTitle.get_text()
		if cbMatchClass.get_active() and entClass.get_text():
			data["wm_class"] = entClass.get_text()
		condition = Condition(**data)

		# Grab selected action
		model, iter = cbProfile.get_model(), cbProfile.get_active_iter()
		action = NoAction()
		if rbProfile.get_active():
			action = ChangeProfileAction(model.get_value(iter, 0))
		elif rbTurnOff.get_active():
			action = TurnOffAction()
		elif rbRestart.get_active():
			action = RestartDaemonAction()

		# Grab & update current row
		model, iter = tvItems.get_selection().get_selected()
		o = model.get_value(iter, 0)
		o.condition = condition
		o.action = action
		model.set_value(iter, 1, condition.describe())
		model.set_value(iter, 2, action.describe(Action.AC_SWITCHER))
		self.hide_dont_destroy(ce)
		self.save_config()

	def on_btAdd_clicked(self, *a):
		"""Handler for "Add Item" button"""
		tvItems = self.builder.get_object("tvItems")
		model = tvItems.get_model()
		o = GObject.GObject()
		o.condition = Condition()
		o.action = NoAction()
		iter = model.append((o, o.condition.describe(), "None"))
		tvItems.get_selection().select_iter(iter)
		self.on_tvItems_cursor_changed()
		self.btEdit_clicked_cb()

	def on_btRemove_clicked(self, *a):
		"""Handler for "Remove Condition" button"""
		tvItems = self.builder.get_object("tvItems")
		model, iter = tvItems.get_selection().get_selected()
		if iter is not None:
			model.remove(iter)
		self.save_config()
		self.on_tvItems_cursor_changed()

	def on_tvItems_cursor_changed(self, *a):
		"""Handles moving cursor in Item List.
		Basically just sets Edit Item and Remove Item buttons sensitivity.
		"""
		tvItems = self.builder.get_object("tvItems")
		btEdit = self.builder.get_object("btEdit")
		btRemove = self.builder.get_object("btRemove")

		model, iter = tvItems.get_selection().get_selected()
		btRemove.set_sensitive(iter is not None)
		btEdit.set_sensitive(iter is not None)

	def on_profiles_loaded(self, profiles):
		cb = self.builder.get_object("cbProfile")
		model = cb.get_model()
		model.clear()
		for f in profiles:
			name = f.get_basename()
			if type(name) is not str:
				try:
					name = name.decode("utf-8")
				except Exception:
					continue
			if name.endswith(".mod"):
				continue
			if name.endswith(".sccprofile"):
				name = name[0:-11]
			model.append((name, f, None))

		cb.set_active(0)

	def on_ConditionEditor_key_press_event(self, w, event):
		"""Checks if pressed key was escape and if yes, closes window"""
		if event.keyval == Gdk.KEY_Escape:
			self.hide_dont_destroy(w)

	def on_cbExactTitle_toggled(self, tg):
		# Ensure that 'Match Title' checkbox is checked and only one of
		# 'match exact title' and 'use regexp' is checked
		if tg.get_active():
			cbMatchTitle = self.builder.get_object("cbMatchTitle")
			cbRegExp = self.builder.get_object("cbRegExp")
			cbMatchTitle.set_active(True)
			cbRegExp.set_active(False)

	def on_cbRegExp_toggled(self, tg):
		# Ensure that 'Match Title' checkbox is checked and only one of
		# 'match exact title' and 'use regexp' is checked
		if tg.get_active():
			cbMatchTitle = self.builder.get_object("cbMatchTitle")
			cbExactTitle = self.builder.get_object("cbExactTitle")
			cbMatchTitle.set_active(True)
			cbExactTitle.set_active(False)

	def on_btClearSensX_clicked(self, *a):
		self.builder.get_object("sclSensX").set_value(1.0)

	def on_btClearSensY_clicked(self, *a):
		self.builder.get_object("sclSensY").set_value(1.0)

	def on_sens_value_changed(self, *a):
		if self._recursing:
			return
		s = (self.builder.get_object("sclSensX").get_value(), self.builder.get_object("sclSensY").get_value())

		profile = self._load_osk_profile()
		if s == (1.0, 1.0):
			profile.pads[LEFT] = OSKCursorAction(LEFT)
			profile.pads[RIGHT] = OSKCursorAction(RIGHT)
		else:
			profile.pads[LEFT] = SensitivityModifier(s[0], s[1], OSKCursorAction(LEFT))
			profile.pads[RIGHT] = SensitivityModifier(s[0], s[1], OSKCursorAction(RIGHT))
		self._save_osk_profile(profile)

	def on_entTitle_changed(self, ent):
		cbRegExp = self.builder.get_object("cbRegExp")
		btSave = self.builder.get_object("btSave")
		cbMatchTitle = self.builder.get_object("cbMatchTitle")
		# Ensure that 'Match Title' checkbox is checked if its entry gets text
		if ent.get_text():
			cbMatchTitle.set_active(True)
		if cbRegExp.get_active():
			# If regexp combobox is active, try to compile expression typed
			# in field and don't allow user to save unless expression is valid
			try:
				re.compile(ent.get_text())
			except Exception as e:
				log.error(e)
				btSave.set_sensitive(False)
				return
		btSave.set_sensitive(True)

	def on_cbOSDColorPreset_changed(self, cb):
		theme = cb.get_model().get_value(cb.get_active_iter(), 0)
		if theme in (None, "None"):
			return
		filename = os.path.join(get_share_path(), "osd-styles", theme)
		data = json.loads(open(filename).read())

		# Transfer values from json to config
		for grp in ("osd_colors", "osk_colors"):
			if grp in data:
				for subkey in self.app.config[grp]:
					if subkey in data[grp]:
						self.app.config[grp][subkey] = data[grp][subkey]

		# Save
		self.app.config["osd_color_theme"] = theme
		self.app.save_config()

	def on_cbOSDStyle_changed(self, cb):
		color_keys = self.app.config["osk_colors"].keys() + self.app.config["osd_colors"].keys()
		osd_style = cb.get_model().get_value(cb.get_active_iter(), 0)
		css_file = os.path.join(get_share_path(), "osd-styles", osd_style)
		with open(css_file) as file:
			first_line = file.read().split("\n")[0]
		used_colors = None  # None means "all"
		if "Used colors:" in first_line:
			used_colors = set(first_line.split(":", 1)[1].strip(" */").split(" "))
			if "all" in used_colors:
				used_colors = None  # None means "all"

		for key in color_keys:
			cb = self.builder.get_object("cb%s" % (key,))
			lbl = self.builder.get_object("lbl%s" % (key,))
			if cb:
				cb.set_sensitive((used_colors is None) or (key in used_colors))
			if lbl:
				lbl.set_sensitive((used_colors is None) or (key in used_colors))
		self.app.config["osd_style"] = osd_style
		self.app.save_config()

	@staticmethod
	def _make_mi_instance(index):
		"""Helper method used by on_cbMI_toggled and load_cbMIs"""
		# label,				class, icon, *init_parameters
		label, order, cls, icon, parameter = GlobalSettings.DEFAULT_MENU_OPTIONS[index]
		if cls == MenuItem:
			instance = MenuItem("item_i%s" % (index,), label, parameter, icon=icon)
		elif cls == Submenu:
			instance = Submenu(parameter, label, icon=icon)
		else:
			instance = cls(parameter)
			instance.icon = icon
			instance.label = label
		return instance

	def on_cbMI_toggled(self, widget):
		"""Called when one of 'Default Menu Items' checkboxes is toggled.
		This actually does kind of magic:
		- 1st, default menu file is loaded
		- 2nd, based on widget name, option from DEFAULT_MENU_OPTIONS is
		  selected
		- 3rd, if this option is not present in loaded menu and checkbox is
		  toggled on, option is added
		- (same for option that is present while checkbox was toggled off)
		- 4rd, default menu is saved
		"""
		if self._recursing:
			return
		try:
			data = MenuData.from_fileobj(open(find_menu("Default.menu")), GuiActionParser())
			index = int(widget.get_name().split("_")[-1])
			instance = GlobalSettings._make_mi_instance(index)
		except Exception:
			log.error(traceback.format_exc())
			self._recursing = True
			widget.set_active(not widget.get_active())
			self._recursing = False
			return

		present = instance.describe().strip(" >") in [x.describe().strip(" >") for x in data]
		if bool(present) == bool(widget.get_active()):
			# User requested to add menu item that's already there
			# (or remove one that's not there)
			return

		items = [x for x in data]
		if widget.get_active():
			# Add item to menu
			order = GlobalSettings.DEFAULT_MENU_OPTIONS[index][1]
			pos = 0
			if order == 1:
				# After last separator
				separators = [x for x in items[1:] if isinstance(x, Separator)]
				if len(separators) > 0:
					pos = items.index(separators[-1]) + 1
			elif order == 2:
				# At very end
				pos = len(items)
			if isinstance(instance, MenuGenerator):
				items.insert(pos, Separator(instance.label))
				pos += 1
			items.insert(pos, instance)
		else:
			if isinstance(instance, MenuGenerator):
				items = [x for x in items if not (isinstance(x, Separator) and x.label == instance.label)]
			items = [x for x in items if instance.describe().strip(" >") != x.describe().strip(" >")]

		path = os.path.join(get_menus_path(), "Default.menu")
		data = MenuData(*items)
		jstr = Encoder(sort_keys=True, indent=4).encode(data)
		open(path, "w").write(jstr)
		log.debug("Wrote menu file %s", path)

	def load_cbMIs(self):
		"""See above. This method just parses Default menu and checks
		boxes for present menu items.
		"""
		try:
			data = MenuData.from_fileobj(open(find_menu("Default.menu")))
		except Exception:
			# Shouldn't really happen
			log.error(traceback.format_exc())
			return
		self._recursing = True

		for index in range(len(GlobalSettings.DEFAULT_MENU_OPTIONS)):
			id = "cbMI_%s" % (index,)
			instance = GlobalSettings._make_mi_instance(index)
			present = instance.describe().strip(" >") in [x.describe().strip(" >") for x in data]
			self.builder.get_object(id).set_active(present)

		# cbMI_5, 'Kill Current Window' is special case here. This checkbox
		# should be available only if xdotool utility is installed.
		cbMI_5 = self.builder.get_object("cbMI_5")
		if find_binary("xdotool") == "xdotool":
			# Not found
			cbMI_5.set_sensitive(False)
			cbMI_5.set_tooltip_text(_("Please, install xdotool package to use this feature"))
		else:
			cbMI_5.set_sensitive(True)
			cbMI_5.set_tooltip_text("")

	def on_btAddController_clicked(self, *a):
		from scc.gui.creg.dialog import ControllerRegistration

		cr = ControllerRegistration(self.app)
		cr.window.connect("destroy", self.load_controllers)
		cr.show(self.window)

	def on_btRemoveController_clicked(self, *a):
		tvControllers = self.builder.get_object("tvControllers")
		d = Gtk.MessageDialog(
			parent=self.window,
			flags=Gtk.DialogFlags.MODAL,
			type=Gtk.MessageType.WARNING,
			buttons=Gtk.ButtonsType.YES_NO,
			message_format=_("Unregister controller?"),
		)
		d.format_secondary_text(_("You'll lose all settings for it"))
		if d.run() == -8:
			# Yes
			model, iter = tvControllers.get_selection().get_selected()
			path = model[iter][0]
			try:
				os.unlink(path)
			except Exception as e:
				log.exception(e)
			self._needs_restart()
			self.load_controllers()
		d.destroy()

	def load_controllers(self, *a):
		lstControllers = self.builder.get_object("lstControllers")
		lstControllers.clear()
		devices_path = os.path.join(get_config_path(), "devices")
		if not os.path.exists(devices_path):
			os.makedirs(os.path.join(get_config_path(), "devices"))
		for filename in os.listdir(os.path.join(get_config_path(), "devices")):
			if filename.endswith(".json"):
				if filename.startswith("hid-"):
					drv, usbid, name = filename.split("-", 2)
					name = "%s <i>(%s)</i>" % (name[0:-5], usbid.upper())
				elif "-" in filename:
					drv, name = filename.split("-", 1)
					name = name[0:-5]
				else:
					continue
				path = os.path.join(get_config_path(), "devices", filename)
				lstControllers.append((path, name, self._get_gamepad_icon(drv)))



"""SC-Controller - Action Editor.

Allows to edit button or trigger action.
"""

import logging
from collections import namedtuple

from gi.repository import Gtk

from scc.actions import Action, ButtonAction, NoAction
from scc.constants import SCButtons
from scc.gui.editor import Editor
from scc.macros import Cycle, Macro, PressAction, ReleaseAction, SleepAction
from scc.tools import _

log = logging.getLogger("MacroEditor")


class MacroEditor(Editor):
	GLADE = "macro_editor.glade"

	def __init__(self, app, callback):
		Editor.__init__(self)
		self.app = app
		self.id = None
		self.mode = Action.AC_BUTTON
		self.ac_callback = callback
		self.added_widget = None
		self.setup_widgets()
		self.actions = []

	def update_action_field(self):
		"""Updates field on bottom"""
		entAction = self.builder.get_object("entAction")
		cbMacroType = self.builder.get_object("cbMacroType")
		btAddDelay = self.builder.get_object("btAddDelay")
		entAction.set_text(self._make_action().to_string())

		# Disable all action type comboboxes in Cycle mode,
		# only click is allowed there;
		# Reenable them if action type is set to anything else.
		sens = cbMacroType.get_active() != 2
		for ad in self.actions:
			if isinstance(ad.action, ButtonAction) or isinstance(ad.action, PressAction):
				if sens:
					ad.combo.set_sensitive(True)
				else:
					ad.combo.set_active(0)
					ad.combo.set_sensitive(False)
			elif isinstance(ad.action, SleepAction):
				ad.label.set_sensitive(sens)
				ad.scale.set_sensitive(sens)
		# Do same thing for 'Add Delay' button
		btAddDelay.set_sensitive(sens)

	def _make_action(self):
		"""Generates and returns Action instance"""
		entName = self.builder.get_object("entName")
		cbMacroType = self.builder.get_object("cbMacroType")
		pars = [x[0] for x in self.actions]
		if len(pars) == 0:
			# No action is actually set
			action = NoAction()
		elif cbMacroType.get_active() == 2:
			# Cycle
			pars = filter(lambda a: not isinstance(a, SleepAction), pars)
			action = Cycle(*pars)
		elif cbMacroType.get_active() == 1:
			# Repeating macro
			action = Macro(*pars)
			action.repeat = True
		elif len(pars) == 1:
			# Only one action
			action = pars[0]
		else:
			# Macro
			action = Macro(*pars)
		if entName.get_text().strip() != "":
			action.name = entName.get_text().strip()
		return action

	def _add_action(self, action):
		"""Adds widgets for new action"""
		grActions = self.builder.get_object("grActions")
		model = self.builder.get_object("lstPressClickOrHold")
		i = len(self.actions) + 1
		action.name = None
		action_data = None

		# Buttons
		button_up, button_down, button_clear = Gtk.Button(), Gtk.Button(), Gtk.Button()
		b = Gtk.Button.new_with_label(action.describe(self.mode))

		# Action button
		b.set_property("hexpand", True)
		b.set_property("margin-left", 10)
		b.set_property("margin-right", 10)

		if isinstance(action, ButtonAction) or isinstance(action, PressAction):
			# Combobox
			c = Gtk.ComboBox()
			c.set_model(model)
			c.set_size_request(100, -1)
			renderer = Gtk.CellRendererText()
			c.pack_start(renderer, True)
			c.add_attribute(renderer, "text", 1)
			if isinstance(action, ReleaseAction):
				c.set_active(2)
				b.set_label(action.describe_short())
			elif isinstance(action, PressAction):
				c.set_active(1)
				b.set_label(action.describe_short())
			else:
				c.set_active(0)
			action_data = ActionData(
				action=action,
				button_up=button_up,
				button_down=button_down,
				button_clear=button_clear,
				combo=c,
				button_action=b,
			)

			c.connect("changed", self.on_buttonaction_type_change, i - 1, action_data)
			grActions.attach(c, 0, i, 1, 1)
			grActions.attach(b, 1, i, 1, 1)
		elif isinstance(action, SleepAction):
			# Label and scale
			l = Gtk.Label("")
			l.set_xalign(0.0)
			l.set_size_request(100, -1)
			s = Gtk.Scale.new_with_range(Gtk.Orientation.HORIZONTAL, 5, 5000, 5)
			s.set_draw_value(False)
			s.set_value(action.delay * 1000)
			action_data = ActionData(
				action=action, button_up=button_up, button_down=button_down, button_clear=button_clear, label=l, scale=s,
			)

			s.connect("change_value", self.on_change_delay, action_data)
			grActions.attach(l, 0, i, 1, 1)
			grActions.attach(s, 1, i, 1, 1)
			self.on_change_delay(s, None, action.delay * 1000, action_data)
		else:
			# Placeholder
			l = Gtk.Label("")
			l.set_size_request(100, -1)

			action_data = ActionData(
				action=action,
				button_up=button_up,
				button_down=button_down,
				button_clear=button_clear,
				label=l,
				button_action=b,
			)

			grActions.attach(l, 0, i, 1, 1)
			grActions.attach(b, 1, i, 1, 1)

		b.connect("clicked", self.on_actionb_clicked, i - 1, action_data)
		button_clear.connect("clicked", self.on_clearb_clicked, action_data)
		button_up.connect("clicked", self.on_moveb_clicked, -1, action_data)
		button_down.connect("clicked", self.on_moveb_clicked, 1, action_data)

		# Move Up button
		button_up.set_image(Gtk.Image.new_from_stock("gtk-go-up", Gtk.IconSize.SMALL_TOOLBAR))
		button_up.set_relief(Gtk.ReliefStyle.NONE)

		# Move Down button
		button_down.set_image(Gtk.Image.new_from_stock("gtk-go-down", Gtk.IconSize.SMALL_TOOLBAR))
		button_down.set_relief(Gtk.ReliefStyle.NONE)

		# Clear button
		button_clear.set_image(Gtk.Image.new_from_stock("gtk-delete", Gtk.IconSize.SMALL_TOOLBAR))
		button_clear.set_relief(Gtk.ReliefStyle.NONE)

		# Pack
		grActions.attach(button_up, 2, i, 1, 1)
		grActions.attach(button_down, 3, i, 1, 1)
		grActions.attach(button_clear, 4, i, 1, 1)

		# Disable 'up' button on 1st aciton
		if len(self.actions) == 0:
			button_up.set_sensitive(False)
		# Reenable 'down' button on last action
		if len(self.actions) > 0:
			self.actions[-1].button_down.set_sensitive(True)
		# Disable 'down' on added (now last) action
		button_down.set_sensitive(False)

		self.actions.append(action_data)
		self.update_action_field()
		grActions.show_all()

	def on_moveb_clicked(self, trash, direction, action_data):
		"""Handler for 'move action' buttons"""
		action = action_data.action
		readd = [x.action for x in self.actions]
		index = readd.index(action) + direction
		if index < 0:
			# Not possible to move 1st item up
			return
		if index > len(readd):
			# Not possible to move last item down
			return
		readd.remove(action)
		readd.insert(index, action)
		self._refill_grid(readd)
		self.update_action_field()

	def on_clearb_clicked(self, trash, action_data):
		"""Handler for 'delete action' button"""
		self._clear_grid()
		self.actions.remove(action_data)
		readd = [x.action for x in self.actions]
		self._refill_grid(readd)
		self.update_action_field()

	def on_cbMacroType_changed(self, *a):
		self.update_action_field()

	def on_buttonaction_type_change(self, cb, i, action_data):
		action = action_data.action
		if isinstance(action, (PressAction, ReleaseAction)):
			action = action.action
		if cb.get_active() == 0:
			if isinstance(action, ButtonAction):
				self.actions[i] = action_data._replace(action=ButtonAction(action.button))
		elif cb.get_active() == 1:
			self.actions[i] = action_data._replace(action=PressAction(action))
		else:
			self.actions[i] = action_data._replace(action=ReleaseAction(action))
		self.update_action_field()

	def _clear_grid(self):
		"""Removes everything from UI"""
		grActions = self.builder.get_object("grActions")
		for child in [] + grActions.get_children():
			grActions.remove(child)

	def _refill_grid(self, new_actions):
		"""Removes everything from UI and then adds updated stuff back"""
		self._clear_grid()
		self.actions = []
		for a in new_actions:
			self._add_action(a)

	def on_change_delay(self, scale, trash, value, action_data):
		"""Called when delay slider is moved"""
		ms = int(value)
		action = action_data.action
		label = action_data.label
		action.delay = value / 1000.0
		if ms < 1000:
			label.set_markup(_("<b>Delay: %sms</b>") % (ms,))
		else:
			s = ms / 1000.0
			label.set_markup(_("<b>Delay: %0.2fs</b>") % (s,))
		self.update_action_field()

	def on_actionb_clicked(self, button, i, action_data):
		"""Handler clicking on action name"""

		def on_chosen(id, action):
			readd = [x.action for x in self.actions]
			readd[i] = action
			self._refill_grid(readd)
			self.update_action_field()

		from scc.gui.action_editor import ActionEditor  # Cannot be imported @ top

		ae = ActionEditor(self.app, on_chosen)
		ae.set_title(_("Edit Action"))
		ae.set_input(self.id, action_data.action, mode=self.mode)
		ae.hide_modeshift()
		ae.hide_macro()
		ae.hide_name()
		ae.show(self.window)

	def on_btAddAction_clicked(self, *a):
		"""Handler for Add Action button"""
		self._add_action(NoAction())

	def on_btAddDelay_clicked(self, *a):
		"""Handler for Add Delay button"""
		self._add_action(SleepAction(0.5))

	def on_btClear_clicked(self, *a):
		"""Handler for clear button"""
		action = NoAction()
		if self.ac_callback is not None:
			self.ac_callback(self.id, action)
		self.close()

	def on_btCustomActionEditor_clicked(self, *a):
		"""Handler for 'Custom Editor' button"""
		from scc.gui.action_editor import ActionEditor  # Can't be imported on top

		e = ActionEditor(self.app, self.ac_callback)
		e.set_input(self.id, self._make_action(), mode=self.mode)
		e.hide_action_buttons()
		e.hide_advanced_settings()
		e.set_title(self.window.get_title())
		e.force_page(e.load_component("custom"), True)
		self.send_added_widget(e)
		self.close()
		e.show(self.get_transient_for())

	def on_btOK_clicked(self, *a):
		"""Handler for OK button"""
		a = self._make_action()
		if self.ac_callback is not None:
			self.ac_callback(self.id, a)
		self.close()

	def add_widget(self, label, widget):
		"""See ActionEditor.add_widget
		"""
		lblAddedWidget = self.builder.get_object("lblAddedWidget")
		vbAddedWidget = self.builder.get_object("vbAddedWidget")
		lblAddedWidget.set_label(label)
		lblAddedWidget.set_visible(True)
		for ch in vbAddedWidget.get_children():
			vbAddedWidget.remove(ch)
		self.added_widget = widget
		vbAddedWidget.pack_start(widget, True, False, 0)
		vbAddedWidget.set_visible(True)

	def on_Dialog_destroy(self, *a):
		vbAddedWidget = self.builder.get_object("vbAddedWidget")
		for ch in vbAddedWidget.get_children():
			vbAddedWidget.remove(ch)

	def allow_first_page(self):
		"""For compatibility with action editor. Does nothing"""

	def set_input(self, id, action, mode=Action.AC_BUTTON):
		"""Common part of editor setup"""
		btDefault = self.builder.get_object("btDefault")
		entName = self.builder.get_object("entName")
		cbMacroType = self.builder.get_object("cbMacroType")
		self.id = id
		self.mode = mode
		self.set_title("Macro for %s" % (id.name if id in SCButtons.__members__.values() else str(id),))
		if isinstance(action, Cycle):
			cbMacroType.set_active(2)
		elif action.repeat:
			cbMacroType.set_active(1)
		else:
			cbMacroType.set_active(0)
		for a in action.actions:
			self._add_action(a)
		if action.name is not None:
			entName.set_text(action.name)

	def hide_name(self):
		"""Hides (and clears) name field.
		"""
		self.builder.get_object("lblName").set_visible(False)
		self.builder.get_object("entName").set_visible(False)
		self.builder.get_object("entName").set_text("")


ActionData = namedtuple("ActionData", "action, button_up, button_down, button_clear, button_action,combo, label, scale")
ActionData.__new__.__defaults__ = (None,) * len(ActionData._fields)



"""SC Controller - Steam Controller Wireless Receiver (aka Dongle) Driver.

Called and used when Dongle is detected on USB bus.
Handles one or multiple controllers connected to dongle.
"""

from __future__ import annotations

import logging
import struct
from collections import namedtuple
from enum import IntEnum
from math import cos, sin
from math import pi as PI

from scc.config import Config
from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN, SCButtons
from scc.controller import Controller
from scc.drivers.usb import USBDevice, register_hotplug_device

VENDOR_ID = 0x28DE
PRODUCT_ID = 0x1142
FIRST_ENDPOINT = 2
FIRST_CONTROLIDX = 1
INPUT_FORMAT = [
	("b", "type"),
	("x", "ukn_01"),
	("B", "status"),
	("x", "ukn_02"),
	("H", "seq"),
	("x", "ukn_03"),
	("I", "buttons"),
	("B", "ltrig"),
	("B", "rtrig"),
	("x", "ukn_04"),
	("x", "ukn_05"),
	("x", "ukn_06"),
	("h", "lpad_x"),
	("h", "lpad_y"),
	("h", "rpad_x"),
	("h", "rpad_y"),
	("4x", "ukn_06"),
	("h", "accel_x"),
	("h", "accel_y"),
	("h", "accel_z"),
	("h", "gpitch"),
	("h", "groll"),
	("h", "gyaw"),
	("h", "q1"),
	("h", "q2"),
	("h", "q3"),
	("h", "q4"),
	("16x", "ukn_07"),
]
FORMATS, NAMES = zip(*INPUT_FORMAT)
TUP_FORMAT = "<" + "".join(FORMATS)
ControllerInput = namedtuple("ControllerInput", " ".join([x for x in NAMES if not x.startswith("ukn_")]))
SCI_NULL = ControllerInput._make(struct.unpack("<" + "".join(FORMATS), b"\x00" * 64))
STICKPRESS = 0b1000000000000000000000000000000


log = logging.getLogger("SCDongle")


class Dongle(USBDevice):
	MAX_ENDPOINTS = 4
	_available_serials = set()  # used only is ignore_serials option is enabled

	def __init__(self, device, handle, daemon):
		self.daemon = daemon
		USBDevice.__init__(self, device, handle)

		self.claim_by(klass=3, subclass=0, protocol=0)
		self._controllers = {}
		self._no_serial = []
		for i in range(Dongle.MAX_ENDPOINTS):
			# Steam dongle apparently can do only 4 controllers at once
			self.set_input_interrupt(FIRST_ENDPOINT + i, 64, self._on_input)

	def close(self):
		# Called when dongle is removed
		for c in self._controllers.values():
			self.daemon.remove_controller(c)
		self._controllers = {}
		USBDevice.close(self)

	def _add_controller(self, endpoint):
		"""Called when new controller is detected either by HOTPLUG message or
		by receiving first input event.
		"""
		ccidx = FIRST_CONTROLIDX + endpoint - FIRST_ENDPOINT
		c = SCController(self, ccidx, endpoint)
		c.configure()
		c.read_serial()
		self._controllers[endpoint] = c

	def _on_input(self, endpoint, data):
		tup = ControllerInput._make(struct.unpack(TUP_FORMAT, data))
		if tup.status == SCStatus.HOTPLUG:
			# Most of INPUT_FORMAT doesn't apply here
			# data[4] is the connect flag (2 == connected)
			if data[4] == 2:
				# Controller connected
				if endpoint not in self._controllers:
					self._add_controller(endpoint)
			# Controller disconnected
			elif endpoint in self._controllers:
				self.daemon.remove_controller(self._controllers[endpoint])
				self._controllers[endpoint].disconnected()
				del self._controllers[endpoint]
		elif tup.status == SCStatus.INPUT:
			if endpoint not in self._controllers:
				self._add_controller(endpoint)
			elif len(self._no_serial):
				for x in self._no_serial:
					x.read_serial()
				self._no_serial = []
			else:
				self._controllers[endpoint].input(tup)


class SCStatus(IntEnum):
	IDLE = 0x04
	INPUT = 0x01
	HOTPLUG = 0x03


class SCPacketType(IntEnum):
	OFF = 0x9F
	AUDIO = 0xB6
	CLEAR_MAPPINGS = 0x81
	CONFIGURE = 0x87
	LED = 0x87
	CALIBRATE_JOYSTICK = 0xBF
	CALIBRATE_TRACKPAD = 0xA7
	SET_AUDIO_INDICES = 0xC1
	LIZARD_MODE = 0x8E
	FEEDBACK = 0x8F
	RESET = 0x95
	GET_SERIAL = 0xAE


class SCPacketLength(IntEnum):
	LED = 0x03
	OFF = 0x04
	FEEDBACK = 0x07
	CONFIGURE = 0x15
	CONFIGURE_BT = 0x0F
	GET_SERIAL = 0x15


class SCConfigType(IntEnum):
	LED = 0x2D
	CONFIGURE = 0x32
	CONFIGURE_BT = 0x18


class SCController(Controller):
	def __init__(self, driver: str, ccidx: int, endpoint: int):
		Controller.__init__(self)
		self._driver = driver
		self._endpoint = endpoint
		self._idle_timeout = 600
		self._enable_gyros = False
		self._input_rotation_l = 0
		self._input_rotation_r = 0
		self._led_level = 10
		# TODO: Is serial really used anywhere?
		self._serial = "0000000000"
		self._id = self._generate_id() if driver else "-"
		self._old_state = SCI_NULL
		self._ccidx = ccidx

	def get_type(self) -> str:
		return "sc"

	def __repr__(self) -> str:
		return "<SCWireless %s>" % (self.get_id(),)

	def input(self, idata: ControllerInput) -> None:
		old_state, self._old_state = self._old_state, idata
		if self.mapper:
			# if idata.buttons & SCButtons.LPAD:
			# # STICKPRESS button may signalize pressing stick instead
			# if (idata.buttons & STICKPRESS) and not (idata.buttons & STICKTILT):
			# idata = ControllerInput.replace(buttons=idata.buttons & ~SCButtons.LPAD)

			if self._input_rotation_l or self._input_rotation_r:
				lx, ly = idata.lpad_x, idata.lpad_y
				rx, ry = idata.rpad_x, idata.rpad_y

				if self._input_rotation_l and idata.buttons & SCButtons.LPADTOUCH:
					s, c = sin(self._input_rotation_l), cos(self._input_rotation_l)
					# Adjust LX for rotation and clamp
					value = int(idata.lpad_x * c - idata.lpad_y * s)
					lx = max(STICK_PAD_MIN, min(STICK_PAD_MAX, value))

					# Adjust LY for rotation and clamp
					value = int(idata.lpad_x * s + idata.lpad_y * c)
					ly = max(STICK_PAD_MIN, min(STICK_PAD_MAX, value))

				if self._input_rotation_r and idata.buttons & SCButtons.RPADTOUCH:
					s, c = sin(self._input_rotation_r), cos(self._input_rotation_r)

					# Adjust RX for rotation and clamp
					value = int(idata.rpad_x * c - idata.rpad_y * s)
					rx = max(STICK_PAD_MIN, min(STICK_PAD_MAX, value))

					# Adjust RY for rotation and clamp
					value = int(idata.rpad_x * s + idata.rpad_y * c)
					ry = max(STICK_PAD_MIN, min(STICK_PAD_MAX, value))

				# TODO: This is awfull :(
				idata = ControllerInput(
					idata.type,
					idata.status,
					idata.seq,
					idata.buttons,
					idata.ltrig,
					idata.rtrig,
					lx,
					ly,
					rx,
					ry,
					idata.accel_x,
					idata.accel_y,
					idata.accel_z,
					idata.gpitch,
					idata.groll,
					idata.gyaw,
					idata.q1,
					idata.q2,
					idata.q3,
					idata.q4,
				)

			self.mapper.input(self, old_state, idata)

	def _generate_id(self):
		"""ID is generated as 'scX' where where 'X' starts as 0 and increases
		as more controllers are connected.

		This is used only when reading serial numbers from device is disabled.
		sc_by_cable generates ids in scBUS:PORT format.
		"""
		magic_number = 1
		tp = self.get_type()
		id = None
		while id is None or id in self._driver.daemon.get_active_ids():
			id = "%s%s" % (
				tp,
				magic_number,
			)
			magic_number += 1
		return id

	def read_serial(self):
		"""Requests and reads serial number from controller"""
		if Config()["ignore_serials"]:
			# Special exception for cases when controller drops instead of
			# sending serial number. See issue #103
			self.generate_serial()
			self.on_serial_got()
			return

		def cb(rawserial):
			size, serial = struct.unpack(">xBx12s49x", rawserial)
			if size > 1:
				serial = serial.strip(b" \x00").decode("ASCII")
				self._serial = serial
				self.on_serial_got()
			else:
				self._driver._no_serial.append(self)

		self._driver.make_request(
			self._ccidx, cb, struct.pack(">BBB61x", SCPacketType.GET_SERIAL, SCPacketLength.GET_SERIAL, 0x01),
		)

	def generate_serial(self):
		"""Called only if ignore_serials is enabled"""
		if len(self._driver._available_serials) > 0:
			self._serial = self._driver._available_serials.pop()
		else:
			self._serial = self.get_id()
		log.debug("Not requesting serial number for SC %s", self._serial)

	def on_serial_got(self):
		try:
			log.debug("Got wireless SC with serial %s", self._serial)
		except UnicodeDecodeError:
			log.debug("Failed to decode wireless SC serial")
			self._serial = self._driver._available_serials.pop()
		self._id = str(self._serial)
		self._driver.daemon.add_controller(self)

	def apply_config(self, config: dict):
		self.configure(idle_timeout=int(config["idle_timeout"]), led_level=float(config["led_level"]))
		self._input_rotation_l = float(config["input_rotation_l"]) * PI / -180.0
		self._input_rotation_r = float(config["input_rotation_r"]) * PI / -180.0

	def disconnected(self):
		# If ignore_serials config option is enabled, fake serial used by this
		# controller is stored away and reused when next controller is connected
		if Config()["ignore_serials"]:
			self._driver._available_serials.add(self._serial)

	FORMAT1 = b">BBBBB13sB2s43x"
	# Has to be overriden in sc_by_cable
	FORMAT2 = b">BBBB59x"

	def configure(
		self, idle_timeout: int | None = None, enable_gyros: bool | None = None, led_level: int | None = None,
	):
		"""Sets and, if possible, sends configuration to controller.
		Only value that is provided is changed.
		'idle_timeout' is in seconds.
		'led_level' is precent (0-100)
		"""
		# ------
		"""
		packet format:
		 - uint8_t type - SCPacketType.CONFIGURE
		 - uint8_t size - SCPacketLength.CONFIGURE or SCPacketLength.LED
		 - uint8_t config_type - SCConfigType.CONFIGURE or SCConfigType.LED
		 - 61B data

		Format for data when configuring controller:
		 - uint16	timeout
		 - 13B		unknown1 - (0x18, 0x00, 0x00, 0x31, 0x02, 0x00, 0x08, 0x07, 0x00, 0x07, 0x07, 0x00, 0x30)
		 - uint8	enable gyro sensor - 0x14 enables, 0x00 disables
		 - 2B		unknown2 - (0x00, 0x2e)
		 - 43B		unused

		Format for data when configuring led:
		 - uint8	led
		 - 60B		unused
		"""

		if idle_timeout is not None:
			self._idle_timeout = idle_timeout
		if enable_gyros is not None:
			self._enable_gyros = enable_gyros
		if led_level is not None:
			self._led_level = led_level

		unknown1 = b"\x18\x00\x00\x31\x02\x00\x08\x07\x00\x07\x07\x00\x30"
		unknown2 = b"\x00\x2e"
		timeout1 = self._idle_timeout & 0x00FF
		timeout2 = (self._idle_timeout & 0xFF00) >> 8

		# Timeout & Gyros
		self._driver.overwrite_control(
			self._ccidx,
			struct.pack(
				self.FORMAT1,
				SCPacketType.CONFIGURE,
				SCPacketLength.CONFIGURE,
				SCConfigType.CONFIGURE,
				timeout1,
				timeout2,
				unknown1,
				# 0x10 (Gyro) | 0x08 (Accel) | 0x04 (Quat)
				0x1C if self._enable_gyros else 0,
				unknown2,
			),
		)

		# LED
		self._driver.overwrite_control(
			self._ccidx,
			struct.pack(
				self.FORMAT2, SCPacketType.CONFIGURE, SCPacketLength.LED, SCConfigType.LED, int(self._led_level),
			),
		)

	def set_led_level(self, level: int):
		level = min(100, int(level)) & 0xFF
		if self._led_level != level:
			self._led_level = level
			self._driver.overwrite_control(
				self._ccidx, struct.pack(">BBBB59x", SCPacketType.CONFIGURE, 0x03, SCConfigType.LED, self._led_level),
			)

	def set_gyro_enabled(self, enabled: bool):
		self.configure(enable_gyros=enabled)

	def turnoff(self):
		log.debug("Turning off the controller...")

		# Mercilessly stolen from scraw library
		self._driver.send_control(self._ccidx, struct.pack("<BBBBBB", SCPacketType.OFF, 0x04, 0x6F, 0x66, 0x66, 0x21))

	def get_gyro_enabled(self):
		"""Returns True if gyroscope input is currently enabled"""
		return self._enable_gyros

	def feedback(self, data):
		self._feedback(*data.data)

	def _feedback(self, position, amplitude=128, period=0, count=1):
		"""Add haptic feedback to be send on next usb tick.

		@param int position		haptic to use 1 for left 0 for right
		@param int amplitude	signal amplitude from 0 to 65535
		@param int period		signal period from 0 to 65535
		@param int count		number of period to play
		"""
		if amplitude >= 0:
			self._driver.send_control(
				self._ccidx, struct.pack("<BBBHHH", SCPacketType.FEEDBACK, 0x07, position, amplitude, period, count),
			)


def init(daemon, config: dict) -> Dongle | bool:
	"""Registers hotplug callback for controller dongle"""

	def cb(device, handle):
		return Dongle(device, handle, daemon)

	register_hotplug_device(cb, VENDOR_ID, PRODUCT_ID)
	return True



#!/usr/bin/env python3
"""SC-Controller - Action Editor - Button Component

Assigns emulated button to physical button
"""

import logging

from gi.repository import Gtk

from scc.actions import Action, AxisAction, ButtonAction, MouseAction, MultiAction, NoAction
from scc.gui.ae import AEComponent
from scc.gui.area_to_action import action_to_area
from scc.gui.chooser import Chooser
from scc.gui.key_grabber import KeyGrabber
from scc.gui.parser import InvalidAction
from scc.macros import Cycle, Macro, PressAction, ReleaseAction
from scc.tools import _
from scc.uinput import Keys, Rels

log = logging.getLogger("AE.Buttons")

__all__ = ["ButtonsComponent"]


class ButtonsComponent(AEComponent, Chooser):
	GLADE = "ae/buttons.glade"
	NAME = "buttons"
	IMAGES = {"buttons": "buttons.svg"}
	CTXS = Action.AC_BUTTON | Action.AC_MENU
	PRIORITY = 1
	MODIFIER_KEYS = (
		Keys.KEY_LEFTSHIFT,
		Keys.KEY_LEFTMETA,
		Keys.KEY_LEFTALT,
		Keys.KEY_LEFTCTRL,
		Keys.KEY_RIGHTMETA,
		Keys.KEY_RIGHTSHIFT,
		Keys.KEY_RIGHTCTRL,
		Keys.KEY_RIGHTALT,
	)

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		Chooser.__init__(self, app)
		self.axes_allowed = True
		self.keys = set()

	def load(self):
		if not self.loaded:
			AEComponent.load(self)
			self.setup_image()
			if self.app.osd_mode:
				self.builder.get_object("btnGrabKey").set_sensitive(False)
				self.builder.get_object("btnGrabAnother").set_sensitive(False)

	def area_action_selected(self, area, action):
		self.set_active_area(area)
		self.editor.set_action(action)

	def set_action(self, mode, action):
		cbToggle = self.builder.get_object("cbToggle")
		cbRepeat = self.builder.get_object("cbRepeat")
		if self.handles(mode, action):
			self.keys = set()
			is_togle, is_repeat = False, False
			if isinstance(action, MultiAction):
				for a in action.actions:
					if isinstance(a, ButtonAction):
						self.keys.add(a.button)
			elif isinstance(action, ButtonAction):
				self.keys.add(action.button)
			elif isinstance(action, Macro):
				# Macro goes here only if it is button repeat
				self.keys.add(action.actions[0].button)
				is_repeat = True
			elif isinstance(action, Cycle):
				# There is only one case when self.handles returns True for Cycle
				self.keys.add(action.actions[0].action.button)
				is_togle = True
			cbToggle.set_active(is_togle)
			cbRepeat.set_active(is_repeat)
			area = action_to_area(action)
			if area is not None:
				self.set_active_area(area)
				return
		self.set_active_area(None)

	def get_button_title(self):
		return _("Key or Button")

	def handles(self, mode, action):
		# Handles ButtonAction and MultiAction if all subactions are ButtonAction
		if isinstance(action, (ButtonAction, NoAction, InvalidAction)):
			return True
		if isinstance(action, AxisAction):
			return len(action.parameters) == 1
		if isinstance(action, MouseAction):
			if action.get_axis() == Rels.REL_WHEEL:
				return True
		if is_button_togle(action):
			return True
		if is_button_repeat(action):
			return True
		if isinstance(action, MultiAction):
			if len(action.actions) > 0:
				for a in action.actions:
					if not isinstance(a, ButtonAction):
						return False
				return True
		return False

	def on_key_grabbed(self, keys):
		"""Handles selecting key using "Grab the Key" dialog"""
		self.keys = set(keys)
		self.apply_keys()

	def on_additional_key_grabbed(self, keys):
		self.keys.update(keys)
		self.apply_keys()

	@staticmethod
	def modifiers_first(key):
		if key in ButtonsComponent.MODIFIER_KEYS:
			return 0
		return 1

	def apply_keys(self, *a):
		"""Common part of on_*key_grabbed"""
		cbToggle = self.builder.get_object("cbToggle")
		cbRepeat = self.builder.get_object("cbRepeat")
		keys = sorted(self.keys, key=ButtonsComponent.modifiers_first)
		action = ButtonAction(keys[0])
		if len(keys) > 1:
			actions = [ButtonAction(k) for k in keys]
			action = MultiAction(*actions)
		if cbRepeat.get_active():
			action = Macro(action)
			action.repeat = True
		elif cbToggle.get_active():
			action = Cycle(PressAction(action), ReleaseAction(action))
		self.editor.set_action(action)

	def on_btnGrabKey_clicked(self, *a):
		"""Called when user clicks on 'Grab a Key' button.
		Displays additional dialog.
		"""
		kg = KeyGrabber(self.app)
		kg.grab(self.editor.window, self.editor._action, self.on_key_grabbed)

	def on_btnGrabAnother_clicked(self, *a):
		"""Same as above, but adds another key to action
		"""
		kg = KeyGrabber(self.app)
		kg.grab(self.editor.window, self.editor._action, self.on_additional_key_grabbed)

	def on_cbToggle_toggled(self, cbToggle):
		cbRepeat = self.builder.get_object("cbRepeat")
		if cbToggle.get_active() and cbRepeat.get_active():
			cbRepeat.set_active(False)
		self.apply_keys()

	def on_cbRepeat_toggled(self, cbRepeat):
		cbToggle = self.builder.get_object("cbToggle")
		if cbToggle.get_active() and cbRepeat.get_active():
			cbToggle.set_active(False)
		self.apply_keys()

	def hide_toggle(self):
		"""Hides 'set as toggle button' option"""
		cbToggle = self.builder.get_object("cbToggle")
		btnGrabAnother = self.builder.get_object("btnGrabAnother")
		replacement = Gtk.Label("")
		replacement.set_size_request(*cbToggle.get_size_request())
		btnGrabAnother.set_visible(False)
		parent = cbToggle.get_parent()
		parent.remove(cbToggle)
		parent.pack_start(replacement, True, True, 0)
		parent.reorder_child(replacement, 0)
		replacement.set_visible(True)

	def hide_axes(self):
		"""Prevents user from selecting axes"""
		self.axes_allowed = False


def is_button_togle(action):
	if not isinstance(action, Cycle):
		return False
	if len(action.actions) != 2:
		return False
	if isinstance(action.actions[0], PressAction):
		if isinstance(action.actions[1], ReleaseAction):
			if isinstance(action.actions[0].action, ButtonAction):
				if isinstance(action.actions[1].action, ButtonAction):
					return action.actions[0].action.button == action.actions[1].action.button
	return False


def is_button_repeat(action):
	if isinstance(action, Cycle):
		return False
	if isinstance(action, Macro) and action.repeat:
		if len(action.actions) == 1:
			return isinstance(action.actions[0], ButtonAction)
	return False



#!/usr/bin/env python3
"""SC-Controller - Action Editor

Allows to edit button or trigger action.
"""

import logging

from scc.actions import Action, MultiAction, NoAction, RingAction
from scc.gui.dwsnc import headerbar
from scc.gui.editor import ComboSetter, Editor
from scc.modifiers import ModeModifier
from scc.tools import _

log = logging.getLogger("RingEditor")


class RingEditor(Editor, ComboSetter):
	GLADE = "ring_editor.glade"

	def __init__(self, app, callback):
		Editor.__init__(self)
		self.app = app
		self.id = None
		self.mode = Action.AC_BUTTON
		self.ac_callback = callback
		self.radius = 0.5
		self.actions = [NoAction(), NoAction()]
		self.setup_widgets()

	def setup_widgets(self):
		Editor.setup_widgets(self)
		b = lambda a: self.builder.get_object(a)
		self.action_widgets = (
			# Order goes: Action Button, Clear Button
			(b("btInner"), b("btClearInner")),
			(b("btOuter"), b("btClearOuter")),
		)
		headerbar(self.builder.get_object("header"))

	@staticmethod
	def is_ring_action(obj):
		"""Returns True if object is instance of RingAction or object
		is MultiAction with RingAction as 2nd item
		"""
		if isinstance(obj, RingAction):
			return True
		if isinstance(obj, MultiAction):
			if len(obj.actions) >= 2:
				if isinstance(obj.actions[1], RingAction):
					return True
		return False

	def on_adjRadius_value_changed(self, scale, *a):
		self.radius = scale.get_value()

	def on_sclRadius_format_value(self, scale, value):
		return "%s%%" % (int(value * 100),)

	def on_Dialog_destroy(self, *a):
		self.remove_added_widget()

	def on_btClearRadius_clicked(self, *a):
		self.radius = 0.5
		self._update()

	def on_cbMode_changed(self, cb):
		lblRadius = self.builder.get_object("lblRadius")
		lblInner = self.builder.get_object("lblInner")
		lblOuter = self.builder.get_object("lblOuter")
		key = cb.get_model().get_value(cb.get_active_iter(), 0)
		if key == "inner":
			lblRadius.set_label(_("Inner Ring Radius"))
			lblInner.set_label(_("Inner Ring"))
			lblOuter.set_label(_("Entire Pad"))
		elif key == "outer":
			lblRadius.set_label(_("Outer Ring Starts at"))
			lblInner.set_label(_("Entire Pad"))
			lblOuter.set_label(_("Outer Ring"))
		elif key == "two":
			lblRadius.set_label(_("Inner Ring Radius"))
			lblInner.set_label(_("Inner Ring"))
			lblOuter.set_label(_("Outer Ring"))

	def _choose_editor(self, action, cb):
		if isinstance(action, ModeModifier):
			from scc.gui.modeshift_editor import ModeshiftEditor  # Cannot be imported @ top

			e = ModeshiftEditor(self.app, cb)
			e.set_title(_("Edit Action"))
		else:
			from scc.gui.action_editor import ActionEditor  # Cannot be imported @ top

			e = ActionEditor(self.app, cb)
			e.set_title(_("Edit Action"))
			e.hide_macro()
			e.hide_ring()
		return e

	def on_actionb_clicked(self, clicked_button):
		for i in range(len(self.action_widgets)):
			button, clearb = self.action_widgets[i]
			if button == clicked_button:

				def on_chosen(id, action):
					self.actions[i] = action
					self._update()

				ae = self._choose_editor(self.actions[i], on_chosen)
				ae.set_input(self.id, self.actions[i])
				ae.show(self.window)
				return

	def on_clearb_clicked(self, clicked_button):
		for i in range(len(self.action_widgets)):
			button, clearb = self.action_widgets[i]
			if clearb == clicked_button:
				self.actions[i] = NoAction()
				self._update()
				return

	def on_btClear_clicked(self, *a):
		"""Handler for clear button"""
		action = NoAction()
		if self.ac_callback is not None:
			self.ac_callback(self.id, action)
		self.close()

	def on_btCustomActionEditor_clicked(self, *a):
		"""Handler for 'Custom Editor' button"""
		from scc.gui.action_editor import ActionEditor  # Can't be imported on top

		e = ActionEditor(self.app, self.ac_callback)
		e.set_input(self.id, self._make_action(), mode=self.mode)
		e.hide_action_buttons()
		e.hide_advanced_settings()
		e.set_title(_("Custom Action"))
		e.force_page(e.load_component("custom"), True)
		self.send_added_widget(e)
		self.close()
		e.show(self.get_transient_for())

	def on_btOK_clicked(self, *a):
		"""Handler for OK button"""
		if self.ac_callback is not None:
			self.ac_callback(self.id, self._make_action())
		self.close()

	def _make_action(self):
		"""Generates and returns Action instance"""
		cbMode = self.builder.get_object("cbMode")
		key = cbMode.get_model().get_value(cbMode.get_active_iter(), 0)
		if key == "inner":
			return MultiAction(self.actions[1], RingAction(self.radius, self.actions[0], NoAction()))
		if key == "outer":
			return MultiAction(self.actions[0], RingAction(self.radius, NoAction(), self.actions[1]))
		return RingAction(self.radius, *self.actions)

	def _update(self):
		for i in range(len(self.action_widgets)):
			button, clearb = self.action_widgets[i]
			button.set_label(self.actions[i].describe(Action.AC_BUTTON))
		self.builder.get_object("sclRadius").set_value(self.radius)

	def allow_first_page(self):
		"""For compatibility with action editor. Does nothing"""

	def set_input(self, id, action, mode=None):
		btDefault = self.builder.get_object("btDefault")
		lblPressAlone = self.builder.get_object("lblPressAlone")
		cbMode = self.builder.get_object("cbMode")
		self.id = id

		if isinstance(action, RingAction):
			self.radius = action.radius
			self.actions = [action.inner, action.outer]
			self.set_cb(cbMode, "two")
		elif RingEditor.is_ring_action(action):
			# Goes here only if action is MultiAciton with RingAction as 2nd item
			ring = action.actions[1]
			self.radius = ring.radius
			if ring.inner:
				self.actions = [ring.inner, action.actions[0]]
				self.set_cb(cbMode, "inner")
			else:
				self.actions = [action.actions[0], ring.outer]
				self.set_cb(cbMode, "outer")
		self._update()



import ctypes
import fcntl
from typing import NamedTuple

import ioctl_opt

# input.h
BUS_USB = 0x03
BUS_HIL = 0x04
BUS_BLUETOOTH = 0x05
BUS_VIRTUAL = 0x06

# hid.h
_HID_MAX_DESCRIPTOR_SIZE = 4096


# hidraw.h
class _hidraw_report_descriptor(ctypes.Structure):
	_fields_ = [
		("size", ctypes.c_uint),
		("value", ctypes.c_ubyte * _HID_MAX_DESCRIPTOR_SIZE),
	]


class _hidraw_devinfo(ctypes.Structure):
	_fields_ = [
		("bustype", ctypes.c_uint),
		("vendor", ctypes.c_short),
		("product", ctypes.c_short),
	]


_HIDIOCGRDESCSIZE = ioctl_opt.IOR(ord("H"), 0x01, ctypes.c_int)
_HIDIOCGRDESC = ioctl_opt.IOR(ord("H"), 0x02, _hidraw_report_descriptor)
_HIDIOCGRAWINFO = ioctl_opt.IOR(ord("H"), 0x03, _hidraw_devinfo)
_HIDIOCGRAWNAME = lambda len: ioctl_opt.IOC(ioctl_opt.IOC_READ, ord("H"), 0x04, len)
_HIDIOCGRAWPHYS = lambda len: ioctl_opt.IOC(ioctl_opt.IOC_READ, ord("H"), 0x05, len)
_HIDIOCSFEATURE = lambda len: ioctl_opt.IOC(ioctl_opt.IOC_WRITE | ioctl_opt.IOC_READ, ord("H"), 0x06, len)
_HIDIOCGFEATURE = lambda len: ioctl_opt.IOC(ioctl_opt.IOC_WRITE | ioctl_opt.IOC_READ, ord("H"), 0x07, len)

HIDRAW_FIRST_MINOR = 0
HIDRAW_MAX_DEVICES = 64
HIDRAW_BUFFER_SIZE = 64


class DevInfo(NamedTuple):
	"""Device Info.

	- bustype: one of BUS_USB, BUS_HIL, BUS_BLUETOOTH or BUS_VIRTUAL
	- vendor: device's vendor number
	- product: device's product number
	"""

	bustype: int
	vendor: int
	product: int


class HIDRaw:
	"""Provide methods to access hidraw device's ioctls."""

	def __init__(self, device):
		"""Device (file, fileno).

		A file object or a fileno of an open hidraw device node.
		"""
		self._device = device

	def _ioctl(self, func, arg, mutate_flag: bool = False):
		result = fcntl.ioctl(self._device, func, arg, mutate_flag)
		if result < 0:
			raise OSError(result)

	def read(self, size):
		return self._device.read(size)

	def write(self, buf):
		return self._device.write(buf)

	def getRawReportDescriptor(self) -> str:
		"""Return a binary string containing the raw HID report descriptor."""
		descriptor = _hidraw_report_descriptor()
		size = ctypes.c_uint()
		self._ioctl(_HIDIOCGRDESCSIZE, size, True)
		descriptor.size = size
		self._ioctl(_HIDIOCGRDESC, descriptor, True)
		return "".join(chr(x) for x in descriptor.value[: size.value])

	# TODO: decode descriptor into a python object
	# def getReportDescriptor(self):

	def getInfo(self) -> DevInfo:
		"""Return a DevInfo instance."""
		devinfo = _hidraw_devinfo()
		self._ioctl(_HIDIOCGRAWINFO, devinfo, True)
		return DevInfo(devinfo.bustype, devinfo.vendor, devinfo.product)

	def getName(self, length: int = 512) -> str:
		"""Return device name as an unicode object."""
		name = ctypes.create_string_buffer(length)
		self._ioctl(_HIDIOCGRAWNAME(length), name, True)
		return name.value.decode("UTF-8")

	def getPhysicalAddress(self, length: int = 512) -> str:
		"""Return device's physical address as a string.

		See hidraw documentation for value signification, as it depends on
		device's bus type.
		"""
		name = ctypes.create_string_buffer(length)
		self._ioctl(_HIDIOCGRAWPHYS(length), name, True)
		return name.value

	def sendFeatureReport(self, report, report_num: int = 0) -> None:
		"""Send a feature report."""
		length = len(report) + 1
		buf = bytearray(length)
		buf[0] = report_num
		buf[1:] = report
		self._ioctl(
			_HIDIOCSFEATURE(length),
			(ctypes.c_char * length).from_buffer(buf),
			True,
		)

	def getFeatureReport(self, report_num: int = 0, length: int = 63) -> bytearray:
		"""Receive a feature report.

		Blocks, unless you configured provided file (descriptor) to be
		non-blocking.
		"""
		length += 1
		buf = bytearray(length)
		buf[0] = report_num
		self._ioctl(
			_HIDIOCGFEATURE(length),
			(ctypes.c_char * length).from_buffer(buf),
			True,
		)
		return buf



"""SC Controller - Config

Handles loading, storing and querying config file
"""

import json
import logging
import os

from scc.paths import get_config_path
from scc.profile import Encoder
from scc.special_actions import ChangeProfileAction

log = logging.getLogger("Config")


class Config:
	DEFAULTS = {
		"autoswitch_osd": True,  # True to show OSD message when profile is autoswitched
		"autoswitch": [],  # Empty list of conditions
		"recent_max": 10,  # Number of profiles to keep
		"recent_profiles": [  # Hard-coded list of profiles from default_profiles/
			# This is actually updated by scc-osd-daemon, as that's
			# only thing actually knowing what to put here.
			"Desktop",
			"XBox Controller with High Precision Camera",
			"XBox Controller",
		],
		"drivers": {  # Map of drivers with values of True, Flase
			# or additional driver config where needed.
			# Anything but False means enabled here.
			"sc_dongle": True,
			"sc_by_cable": True,
			"sc_by_bt": True,
			"steamdeck": True,
			"fake": False,  # Used for developement
			"hiddrv": True,
			"evdevdrv": True,
			"ds4drv": True,  # At least one of hiddrv or evdevdrv has to be enabled as well
			"ds5drv": True,
		},
		"fix_xinput": True,  # If True, attempt is done to deatach emulated controller
		# from 'Virtual core pointer' core device.
		"gui": {
			# GUI-only settings
			"enable_status_icon": False,
			"minimize_to_status_icon": True,
			"minimize_on_start": False,
			"autokill_daemon": False,
			"news": {
				# Controls "new in this version" message
				"enabled": True,  # if disabled, no querying is done
				"last_version": "0.3.12",  # last version for which message was displayed
			},
		},
		"controllers": {},
		# output - modifies emulated controller
		# Changing this may be usefull, but can break a lot of things
		"output": {
			"vendor": "0x045e",
			"product": "0x028e",
			"version": "0x110",
			"name": "Microsoft X-Box 360 pad",
			"buttons": 11,
			"rumble": True,
			"axes": [
				(-32768, 32767),  # Axes.ABS_X
				(-32768, 32767),  # Axes.ABS_Y
				(-32768, 32767),  # Axes.ABS_RX
				(-32768, 32767),  # Axes.ABS_RY
				(0, 255),  # Axes.ABS_Z
				(0, 255),  # Axes.ABS_RZ
				(-1, 1),  # Axes.ABS_HAT0X
				(-1, 1),  # Axes.ABS_HAT0Y
			],
		},
		# enable_sniffing - If enabled, another program with write access to
		# ~/.config/scc can ask daemon to send notifications about all
		# (or only some) inputs.
		# This enables GUI to display which physical button was pressed to user.
		"enable_sniffing": False,
		# Style and colors used by OSD
		"osd_style": "Classic.gtkstyle.css",
		"osd_colors": {
			"background": "101010",
			"border": "101010",
			"text": "16BF24",
			"menuitem_border": "101010",
			"menuitem_hilight": "202020",
			"menuitem_hilight_text": "16FF26",
			"menuitem_hilight_border": "16FF26",
			"menuseparator": "2e3436",
		},
		# Colors used by on-screen keyboard
		"osk_colors": {
			"hilight": "7A7A7A",
			"pressed": "B0B0B0",
			"button1": "101010",
			"button1_border": "101010",
			"button2": "2e3436",
			"button2_border": "2e3436",
			"text": "16BF24",
		},
		# Colors used by gesture display. Unlike OSD and OSK, these are RGBA
		"gesture_colors": {
			"background": "160c00ff",
			"grid": "004000ff",
			"line": "ffffff1a",
		},
		# TODO: Config for opacity
		"windows_opacity": 0.95,
		# See drivers/sc_dongle.py, read_serial method
		"ignore_serials": True,
	}

	CONTROLLER_DEFAULTS: dict[str, None | int | str] = {
		# Defaults for controller config
		"name": None,  # Filled with controller ID on runtime
		"icon": None,  # Determined by magic by UI
		"led_level": 80,  # range 0 to 100
		"idle_timeout": 600,  # in seconds, range from 1 to 32767
		"osd_alignment": 0,  # not used yet
		"input_rotation_l": 20,  # range -180 to 180
		"input_rotation_r": -20,  # range -180 to 180
		"menu_control": "STICK",
		"menu_confirm": "A",
		"menu_cancel": "B",
	}

	def __init__(self):
		self.filename: str = os.path.join(get_config_path(), "config.json")
		self.reload()

	def reload(self):
		"""(Re)loads configuration. Works as load(), but handles exceptions"""
		try:
			self.load()
		except Exception as e:
			log.warning("Failed to load configuration; Creating new one.")
			log.warning("Reason: %s", (e,))
			self.create()
		if self.check_values():
			self.save()

	def _check_dict(self, values, defaults):
		"""Recursively checks if 'config' contains all keys in 'defaults'.
		Creates keys with default values where missing.

		Returns True if anything was changed.
		"""
		rv = False
		for d in defaults:
			if d not in values:
				values[d] = defaults[d]
				rv = True
			if type(values[d]) == dict:
				rv = self._check_dict(values[d], defaults[d]) or rv
		return rv

	def check_values(self):
		"""Check if all required values are in place and fill by default
		whatever is missing.

		Returns True if anything gets changed.
		"""
		rv = self._check_dict(self.values, self.DEFAULTS)
		# Special check for autoswitcher after v0.2.17
		if "autoswitch" in self.values:
			for a in self.values["autoswitch"]:
				if "profile" in a:
					a["action"] = ChangeProfileAction(str(a["profile"])).to_string()
					del a["profile"]
					rv = True
		return rv

	def get_controller_config(self, controller_id):
		"""Returns self['controllers'][controller_id], creating new node populated
		with defaults if there is none.
		"""
		if controller_id in self.values["controllers"]:
			# Check values in existing config
			rv = self.values["controllers"][controller_id]
			for key in self.CONTROLLER_DEFAULTS:
				if key not in rv:
					if key in ("input_rotation_l", "input_rotation_r"):
						# Special case, just to not change behavior for existing users
						rv[key] = 0
					else:
						rv[key] = self.CONTROLLER_DEFAULTS[key]
			return rv
		# Create new config
		rv = self.values["controllers"][controller_id] = {
			key: self.CONTROLLER_DEFAULTS[key] for key in self.CONTROLLER_DEFAULTS
		}
		rv["name"] = controller_id
		return rv

	def load(self):
		self.values = json.loads(open(self.filename).read())

	def create(self):
		"""Creates new, empty configuration"""
		self.values = {}
		self.check_values()
		self.save()

	def save(self):
		"""Saves configuration file"""
		# Check & create directory
		if not os.path.exists(get_config_path()):
			os.makedirs(get_config_path())
		# Save
		data = {k: self.values[k] for k in self.values}
		jstr = Encoder(sort_keys=True, indent=4).encode(data)
		open(self.filename, "w").write(jstr)
		log.debug("Configuration saved")

	def __iter__(self):
		for k in self.values:
			yield k

	def get(self, key, default=None):
		return self.values.get(key, default)

	def set(self, key, value):
		self.values[key] = value

	__getitem__ = get
	__setitem__ = set

	def __contains__(self, key):
		"""Returns true if there is such value"""
		return key in self.values



"""SC Controller - Action Editor - Axis Component

Assigns emulated axis to trigger
"""

import logging
import os
from ctypes import POINTER, cast

from gi.repository import Gdk, GdkX11, GLib

from scc.actions import (
	Action,
	AreaAction,
	AxisAction,
	ButtonAction,
	MouseAction,
	MultiAction,
	NoAction,
	RelAreaAction,
	RelWinAreaAction,
	RelXYAction,
	WinAreaAction,
	XYAction,
)
from scc.constants import SCButtons
from scc.gui.ae import AEComponent
from scc.gui.controller_widget import STICKS
from scc.gui.parser import GuiActionParser, InvalidAction
from scc.gui.simple_chooser import SimpleChooser
from scc.lib import xwrappers as X
from scc.modifiers import BallModifier, CircularModifier
from scc.osd.area import Area
from scc.osd.timermanager import TimerManager
from scc.tools import _
from scc.uinput import Axes, Keys, Rels

log = logging.getLogger("AE.AxisAction")

__all__ = ["AxisActionComponent"]


class AxisActionComponent(AEComponent, TimerManager):
	GLADE = "ae/axis_action.glade"
	NAME = "axis_action"
	CTXS = Action.AC_STICK | Action.AC_PAD
	PRIORITY = 3

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		TimerManager.__init__(self)
		self._recursing = False
		self.relative_area = False
		self.osd_area_instance = None
		self.on_wayland = False
		self.circular_axis = MouseAction(Rels.REL_WHEEL)
		self.circular_buttons = [None, None]
		self.button = None
		self.parser = GuiActionParser()

	def load(self):
		if self.loaded:
			return
		AEComponent.load(self)
		cbAreaType = self.builder.get_object("cbAreaType")
		cbAreaType.set_row_separator_func(lambda model, iter: model.get_value(iter, 0) == "-")
		self.on_wayland = "WAYLAND_DISPLAY" in os.environ or not isinstance(
			Gdk.Display.get_default(), GdkX11.X11Display,
		)
		if self.on_wayland:
			self.builder.get_object("lblArea").set_text(
				_("Note: Mouse Region option is not available with Wayland-based display server"),
			)
			self.builder.get_object("grArea").set_sensitive(False)

		# Remove options that are not applicable to currently editted input
		if self.editor.get_id() in STICKS:
			# Remove "Mouse Region", "Mouse" and "Mouse (Emulate Stick)" options
			# when editing stick bindings
			cb = self.builder.get_object("cbAxisOutput")
			for row in cb.get_model():
				if row[2] in ("wheel_pad", "area", "mouse", "mouse_pad"):
					cb.get_model().remove(row.iter)
		else:
			# Remove "Mouse" option when editing pads
			# (it's effectivelly same as Trackpad)
			cb = self.builder.get_object("cbAxisOutput")
			for row in cb.get_model():
				if row[2] in ("wheel_stick", "mouse_stick"):
					cb.get_model().remove(row.iter)

	def hidden(self):
		self.update_osd_area(None)

	def set_action(self, mode, action):
		if self.handles(mode, action):
			cb = self.builder.get_object("cbAxisOutput")
			if isinstance(action, AreaAction):
				self.load_area_action(action)
				self.set_cb(cb, "area", 2)
				self.update_osd_area(action)
				return
			self.update_osd_area(None)
			if isinstance(action, MouseAction):
				self.load_mouse_action(action)
			elif isinstance(action, CircularModifier):
				self.load_circular_action(action)
			elif isinstance(action, ButtonAction):
				self.load_button_action(action)
			elif isinstance(action, XYAction):
				if self.editor.friction > 0:
					self.load_mouse_action(action)
				else:
					p = [None, None]
					for x in (0, 1):
						if len(action.actions[0].strip().parameters) >= x:
							if len(action.actions[x].strip().parameters) > 0:
								p[x] = action.actions[x].strip().parameters[0]
					if p[0] == Axes.ABS_X and p[1] == Axes.ABS_Y:
						self.set_cb(cb, "lstick", 2)
					elif p[0] == Axes.ABS_RX and p[1] == Axes.ABS_RY:
						if isinstance(action, RelXYAction):
							self.set_cb(cb, "rstick_rel", 2)
						else:
							self.set_cb(cb, "rstick", 2)
					elif p[0] == Rels.REL_HWHEEL and p[1] == Rels.REL_WHEEL:
						self.set_cb(cb, "wheel_pad", 2)
						self.set_cb(cb, "wheel_stick", 2)
			else:
				self.set_cb(cb, "none", 2)

	def update_osd_area(self, action):
		"""Updates preview area displayed on screen"""
		if action:
			if self.osd_area_instance is None:
				if self.on_wayland:
					# Cannot display preview with non-X11 backends
					return
				self.osd_area_instance = Area()
				self.osd_area_instance.show()
			action.update_osd_area(self.osd_area_instance, FakeMapper(self.editor))
			self.timer("area", 0.5, self.update_osd_area, action)
		elif self.osd_area_instance:
			self.osd_area_instance.quit()
			self.osd_area_instance = None
			self.cancel_timer("area")

	def load_circular_action(self, action):
		cbAxisOutput = self.builder.get_object("cbAxisOutput")
		btCircularAxis = self.builder.get_object("btCircularAxis")
		btCircularButton0 = self.builder.get_object("btCircularButton0")
		btCircularButton1 = self.builder.get_object("btCircularButton1")

		# Turn action into list of subactions (even if it's just single action)
		if isinstance(action.action, MultiAction):
			actions = action.action.actions
		else:
			actions = [action.action]

		# Parse that list
		self.circular_axis, self.circular_buttons = NoAction(), [None, None]
		for action in actions:
			if isinstance(action, ButtonAction):
				self.circular_buttons = [action.button, action.button2]
			else:
				self.circular_axis = action

		# Set labels
		b0, b1 = self.circular_buttons
		btCircularButton0.set_label(ButtonAction.describe_button(b0))
		btCircularButton1.set_label(ButtonAction.describe_button(b1))
		btCircularAxis.set_label(self.circular_axis.describe(Action.AC_PAD))

		self.set_cb(cbAxisOutput, "circular", 2)

	def load_button_action(self, action):
		self.button = action
		cbAxisOutput = self.builder.get_object("cbAxisOutput")
		btSingleButton = self.builder.get_object("btSingleButton")
		btSingleButton.set_label(self.button.describe(Action.AC_PAD))
		self.set_cb(cbAxisOutput, "button", 2)

	def load_mouse_action(self, action):
		cbMouseOutput = self.builder.get_object("cbMouseOutput")
		cbAxisOutput = self.builder.get_object("cbAxisOutput")
		self._recursing = True
		if isinstance(action, MouseAction):
			if self.editor.get_id() in STICKS:
				self.set_cb(cbMouseOutput, "mouse_stick", 1)
				self.set_cb(cbAxisOutput, "mouse_stick", 2)
			else:
				self.set_cb(cbMouseOutput, "mouse", 1)
				self.set_cb(cbAxisOutput, "mouse", 2)
		elif isinstance(action, XYAction):
			if isinstance(action.x, AxisAction):
				if action.x.parameters[0] == Axes.ABS_X:
					self.set_cb(cbMouseOutput, "left", 1)
				else:
					self.set_cb(cbMouseOutput, "right", 1)
				self.set_cb(cbAxisOutput, "mouse", 2)
			elif isinstance(action.x, MouseAction):
				if self.editor.get_id() in STICKS:
					self.set_cb(cbAxisOutput, "wheel_stick", 2)
				else:
					self.set_cb(cbAxisOutput, "wheel_pad", 2)
		self._recursing = False

	def load_area_action(self, action):
		"""Load AreaAction values into UI.
		"""
		cbAreaType = self.builder.get_object("cbAreaType")

		x1, y1, x2, y2 = action.coords
		self.relative_area = False
		if isinstance(action, RelAreaAction):
			key = "screensize"
			self.relative_area = True
			x1, y1, x2, y2 = x1 * 100.0, y1 * 100.0, x2 * 100.0, y2 * 100.0
		elif isinstance(action, RelWinAreaAction):
			key = "windowsize"
			self.relative_area = True
			x1, y1, x2, y2 = x1 * 100.0, y1 * 100.0, x2 * 100.0, y2 * 100.0
		else:
			t1 = "1" if x1 < 0 and x2 < 0 else "0"
			t2 = "1" if y1 < 0 and y2 < 0 else "0"
			x1, y1, x2, y2 = abs(x1), abs(y1), abs(x2), abs(y2)
			if x2 < x1:
				x1, x2 = x2, x1
			if y2 < y1:
				y1, y2 = y2, y1
			if isinstance(action, WinAreaAction):
				key = "window-%s%s" % (t1, t2)
			else:
				key = "screen-%s%s" % (t1, t2)

		self._recursing = True
		self.builder.get_object("sbAreaX1").set_value(x1)
		self.builder.get_object("sbAreaY1").set_value(y1)
		self.builder.get_object("sbAreaX2").set_value(x2)
		self.builder.get_object("sbAreaY2").set_value(y2)
		self.builder.get_object("cbAreaOSDEnabled").set_active(self.editor.osd)
		self.builder.get_object("cbAreaClickEnabled").set_active(self.pressing_pad_clicks())
		for row in cbAreaType.get_model():
			if key == row[1]:
				cbAreaType.set_active_iter(row.iter)
				break
		self._recursing = False

	def on_btCircularAxis_clicked(self, *a):
		def cb(action):
			self.circular_axis = action
			btCircularAxis = self.builder.get_object("btCircularAxis")
			btCircularAxis.set_label(action.describe(Action.AC_PAD))
			self.editor.set_action(self.make_circular_action())

		b = SimpleChooser(self.app, "axis", cb)
		b.set_title(_("Select Axis"))
		b.display_action(Action.AC_STICK, self.circular_axis)
		b.show(self.editor.window)

	def on_btCircularButton_clicked(self, button, *a):
		index = 0 if button == self.builder.get_object("btCircularButton0") else 1

		def cb(action):
			self.circular_buttons[index] = action.button
			btCircularButton = self.builder.get_object("btCircularButton%s" % (index,))
			btCircularButton.set_label(action.describe(Action.AC_PAD))
			self.editor.set_action(self.make_circular_action())

		b = SimpleChooser(self.app, "buttons", cb)
		b.set_title(_("Select Button"))
		b.display_action(Action.AC_STICK, self.circular_axis)
		b.show(self.editor.window)

	def on_btClearCircularAxis_clicked(self, *a):
		btCircularAxis = self.builder.get_object("btCircularAxis")
		self.circular_axis = NoAction()
		btCircularAxis.set_label(self.circular_axis.describe(Action.AC_PAD))
		self.editor.set_action(self.make_circular_action())

	def on_btClearCircularButtons_clicked(self, *a):
		btCircularButton0 = self.builder.get_object("btCircularButton0")
		btCircularButton1 = self.builder.get_object("btCircularButton1")
		self.circular_buttons = [None, None]
		btCircularButton0.set_label(NoAction().describe(Action.AC_PAD))
		btCircularButton1.set_label(NoAction().describe(Action.AC_PAD))
		self.editor.set_action(self.make_circular_action())

	def on_btSingleButton_clicked(self, *a):
		def cb(action):
			self.button = action
			btSingleButton = self.builder.get_object("btSingleButton")
			btSingleButton.set_label(self.button.describe(Action.AC_PAD))
			self.editor.set_action(self.button)

		b = SimpleChooser(self.app, "buttons", cb)
		b.set_title(_("Select Button"))
		b.display_action(Action.AC_STICK, self.circular_axis)
		b.show(self.editor.window)

	def on_cbAreaOSDEnabled_toggled(self, *a):
		self.editor.builder.get_object("cbOSD").set_active(self.builder.get_object("cbAreaOSDEnabled").get_active())

	def pressing_pad_clicks(self):
		"""Returns True if currently edited pad is set to press left mouse
		button when pressed.
		(yes, this is used somewhere)
		"""
		side = getattr(SCButtons, self.editor.get_id())
		c_action = self.app.current.buttons[side]
		if isinstance(c_action, ButtonAction):
			return c_action.button == Keys.BTN_LEFT
		return False

	def on_ok(self, action):
		if isinstance(action.strip(), AreaAction):
			# Kinda hacky way to set action on LPAD press or RPAD press
			# when user selects Mouse Area as ouput and checks
			# 'Pressing the Pad Clicks' checkbox
			side = getattr(SCButtons, self.editor.get_id())
			clicks = self.pressing_pad_clicks()

			if self.builder.get_object("cbAreaClickEnabled").get_active():
				if not clicks:
					# Turn pad press into mouse clicks
					self.app.set_action(self.app.current, side, ButtonAction(Keys.BTN_LEFT))
			elif clicks:
				# Clear action created above if checkbox is uncheck
				self.app.set_action(self.app.current, side, NoAction())

	def on_mouse_options_changed(self, *a):
		if self._recursing:
			return
		action = self.make_mouse_action()
		self.editor.set_action(action)

	def make_mouse_action(self):
		"""Loads values from UI into trackball-related action
		"""
		cbMouseOutput = self.builder.get_object("cbMouseOutput")
		a_str = cbMouseOutput.get_model().get_value(cbMouseOutput.get_active_iter(), 2)
		return self.parser.restart(a_str).parse()

	def make_circular_action(self):
		"""Constructs Circular Modifier
		"""
		if self.circular_axis and any(self.circular_buttons):
			return CircularModifier(MultiAction(self.circular_axis, ButtonAction(*self.circular_buttons)))
		if any(self.circular_buttons):
			return CircularModifier(ButtonAction(*self.circular_buttons))
		return CircularModifier(self.circular_axis)

	def make_area_action(self):
		"""Loads values from UI into new AreaAction or subclass.
		"""
		# Prepare
		cbAreaType = self.builder.get_object("cbAreaType")
		# Read numbers
		x1 = self.builder.get_object("sbAreaX1").get_value()
		y1 = self.builder.get_object("sbAreaY1").get_value()
		x2 = self.builder.get_object("sbAreaX2").get_value()
		y2 = self.builder.get_object("sbAreaY2").get_value()
		# Determine exact action type by looking into Area Type checkbox
		# (this part may seem little crazy)
		# ... numbers
		key = cbAreaType.get_model().get_value(cbAreaType.get_active_iter(), 1)
		if "-" in key:
			if key[-2] == "1":
				# Before-last character ius "1", that means that X coords are
				# counted from other side and has to be negated
				x1, x2 = -x1, -x2
			if key[-1] == "1":
				# Key ends with "1". Same thing as above but for Y coordinate
				y1, y2 = -y1, -y2
		if "size" in key:
			x1, y1, x2, y2 = x1 / 100.0, y1 / 100.0, x2 / 100.0, y2 / 100.0
		# ... class
		if "window-" in key:
			cls = WinAreaAction
			self.relative_area = False
		elif key == "screensize":
			cls = RelAreaAction
			self.relative_area = True
		elif key == "windowsize":
			cls = RelWinAreaAction
			self.relative_area = True
		else:  # "screen" in key
			cls = AreaAction
			self.relative_area = False
		if not self.relative_area:
			x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)

		return cls(x1, y1, x2, y2)

	def get_button_title(self):
		return _("Joystick / Mouse")

	def handles(self, mode, action):
		if isinstance(action, (NoAction, MouseAction, CircularModifier, InvalidAction, AreaAction, ButtonAction)):
			return True
		if isinstance(action, BallModifier):
			if isinstance(action.action, XYAction):
				return isinstance(action.action.x, (AxisAction, MouseAction)) and isinstance(
					action.action.x, (AxisAction, MouseAction),
				)
			return isinstance(action.action, MouseAction)
		if isinstance(action, XYAction):
			p = [None, None]
			for x in (0, 1):
				if len(action.actions[0].strip().parameters) >= x:
					if len(action.actions[x].strip().parameters) > 0:
						p[x] = action.actions[x].strip().parameters[0]
			if (p[0] == Axes.ABS_X and p[1] == Axes.ABS_Y) or (p[0] == Axes.ABS_RX and p[1] == Axes.ABS_RY):
				return True
			if (p[0] == Axes.ABS_HAT0X and p[1] == Axes.ABS_HAT0Y) or (p[0] == Rels.REL_HWHEEL and p[1] == Rels.REL_WHEEL):
				return True
		return False

	def on_area_options_changed(self, *a):
		if self._recursing:
			return
		action = self.make_area_action()
		self.editor.set_action(action)
		self.update_osd_area(action)
		for x in ("sbAreaX1", "sbAreaX2", "sbAreaY1", "sbAreaY2"):
			spin = self.builder.get_object(x)
			if self.relative_area:
				spin.get_adjustment().set_upper(100)
			else:
				spin.get_adjustment().set_upper(1000)
			self.on_sbArea_output(spin)

	def on_sbArea_output(self, button, *a):
		if self.relative_area:
			button.set_text("%s %%" % (button.get_value()))
		else:
			button.set_text("%s px" % (int(button.get_value())))

	def on_sbArea_focus_out_event(self, button, *a):
		GLib.idle_add(self.on_sbArea_output, button)

	def on_sbArea_changed(self, button, *a):
		self.on_sbArea_output(button)
		self.on_area_options_changed(button)

	def on_lblSetupTrackball_activate_link(self, trash, link):
		self.editor.on_link(link)

	def on_cbAxisOutput_changed(self, *a):
		cbAxisOutput = self.builder.get_object("cbAxisOutput")
		stActionData = self.builder.get_object("stActionData")
		key = cbAxisOutput.get_model().get_value(cbAxisOutput.get_active_iter(), 2)
		if key == "area":
			stActionData.set_visible_child(self.builder.get_object("grArea"))
			action = self.make_area_action()
			self.update_osd_area(action)
		elif key == "button":
			stActionData.set_visible_child(self.builder.get_object("vbButton"))
			self.button = self.button or ButtonAction(Keys.BTN_GAMEPAD)
			action = self.button
		elif key == "circular":
			stActionData.set_visible_child(self.builder.get_object("grCircular"))
			action = self.make_circular_action()
		elif key == "mouse":
			stActionData.set_visible_child(self.builder.get_object("vbMose"))
			if not self._recursing and self.editor.friction == 0:
				# When switching to mouse, enable trackball by default
				self.editor.friction = 10
			action = self.make_mouse_action()
		else:
			stActionData.set_visible_child(self.builder.get_object("nothing"))
			action = cbAxisOutput.get_model().get_value(cbAxisOutput.get_active_iter(), 0)
			action = self.parser.restart(action).parse()
			self.update_osd_area(None)

		self.editor.set_action(action)


class FakeMapper:
	"""Class that pretends to be mapper used when calling update_osd_area.
	It has two purposes: To provide get_xdisplay() method that does what it says
	and get_active_window() that returns any other window but window that
	belongs to SC-Controller gui application.
	"""

	def __init__(self, editor):
		self._xdisplay = X.Display(hash(GdkX11.x11_get_default_xdisplay()))
		self.editor = editor

	def get_xdisplay(self):
		return self._xdisplay

	def get_current_window(self):
		"""Gets last active window that was not part of SC-Controller.
		Uses _NET_CLIENT_LIST_STACKING property of root window, value provided
		by window manager. If that fail (because of no or not ICCM compilant WM),
		simply returns root window.
		"""
		root = X.get_default_root_window(self._xdisplay)
		NET_WM_WINDOW_TYPE_NORMAL = X.intern_atom(self._xdisplay, bytes("_NET_WM_WINDOW_TYPE_NORMAL", "utf-8"), False)
		my_windows = [x.get_xid() for x in Gdk.Screen.get_default().get_toplevel_windows()]
		nitems, prop = X.get_window_prop(
			self._xdisplay, root, bytes("_NET_CLIENT_LIST_STACKING", "utf-8"), max_size=0x8000,
		)

		if nitems > 0:
			for i in reversed(range(nitems)):
				window = cast(prop, POINTER(X.XID))[i]
				if window in my_windows:
					# skip over my own windows
					continue
				if not X.is_window_visible(self._xdisplay, window):
					# skip minimized and invisible windows
					continue
				tp = X.get_window_prop(self._xdisplay, window, bytes("_NET_WM_WINDOW_TYPE", "utf-8"))[-1]
				if tp is not None:
					tpval = cast(tp, POINTER(X.Atom)).contents.value
					X.free(tp)
					if tpval != NET_WM_WINDOW_TYPE_NORMAL:
						# skip over non-normal windows
						continue
				X.free(prop)
				return window
		if prop is not None:
			X.free(prop)

		# Failed to get property or there is not any usable window
		return root



#!/usr/bin/env python3
"""SC-Controller - RIBar

Infobar wrapped in Revealer, looks better than sounds.
"""

from gi.repository import GLib, GObject, Gtk


class RIBar(Gtk.Revealer):
	"""Infobar wrapped in Revealer

	Signals:
		Everything from Gtk.Revealer, plus:
		close()
			emitted when the user dismisses the info bar
		response(response_id)
			Emitted when an action widget (button) is clicked
	"""

	__gsignals__ = {
		"response": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
		"close": (GObject.SignalFlags.RUN_FIRST, None, ()),
	}

	### Initialization
	def __init__(self, label, message_type=Gtk.MessageType.INFO, infobar=None, *buttons):
		"""... where label can be Gtk.Widget or str and buttons are tuples
		of (Gtk.Button, response_id)
		"""
		# Init
		Gtk.Revealer.__init__(self)
		self._infobar = infobar or Gtk.InfoBar()
		self._values = {}
		self._label = None
		# Icon
		if infobar is None:
			icon_name = "dialog-information"
			if message_type == Gtk.MessageType.ERROR:
				icon_name = "dialog-error"
			elif message_type == Gtk.MessageType.WARNING:
				icon_name = "dialog-warning"
			icon = Gtk.Image()
			icon.set_from_icon_name(icon_name, Gtk.IconSize.DIALOG)
			self._infobar.get_content_area().pack_start(icon, False, False, 1)
			# Label
			if isinstance(label, Gtk.Widget):
				self._infobar.get_content_area().pack_start(label, True, True, 0)
				self._label = label
			else:
				self._label = Gtk.Label()
				self._label.set_size_request(300, -1)
				self._label.set_markup(label)
				self._label.set_alignment(0, 0.5)
				self._label.set_line_wrap(True)
				self._infobar.get_content_area().add(self._label)
		# Buttons
		for button, response_id in buttons:
			self.add_button(button, response_id)
		# Signals
		self._infobar.connect("close", self._cb_close)
		self._infobar.connect("response", self._cb_response)
		# Settings
		self._infobar.set_message_type(message_type)
		if hasattr(self._infobar, "set_show_close_button"):
			# GTK >3.8
			self._infobar.set_show_close_button(True)
		else:
			self.add_button(Gtk.Button("X"), 0)
		self.set_reveal_child(False)
		# Packing
		self.add(self._infobar)
		self.show_all()

	def _cb_close(self, ib):
		self.emit("close")

	def _cb_response(self, ib, response_id):
		self.emit("response", response_id)

	def disable_close_button(self):
		if hasattr(self._infobar, "set_show_close_button"):
			self._infobar.set_show_close_button(False)

	def add_widget(self, widget, expand=False, fill=True):
		self._infobar.get_content_area().pack_start(widget, expand, fill, 1)
		widget.show()

	def add_button(self, button, response_id):
		self._infobar.add_action_widget(button, response_id)
		self._infobar.show_all()

	def get_label(self):
		"""Returns label widget"""
		return self._label

	def close_on_close(self):
		"""Setups revealer so it will be automaticaly closed, removed and
		destroyed when user clicks to any button, including 'X'
		"""
		self.connect("close", self.close)
		self.connect("response", self.close)

	def close(self, *a):
		"""Closes revealer (with animation), removes it from parent and
		calls destroy()
		"""
		self.set_reveal_child(False)
		GLib.timeout_add(self.get_transition_duration() + 50, self._cb_destroy)

	def _cb_destroy(self, *a):
		"""Callback used by _cb_close method"""
		if self.get_parent() is not None:
			self.get_parent().remove(self)
		self.destroy()

	def set_value(self, key, value):
		"""Stores some metadata"""
		self._values[key] = value

	def get_value(self, key):
		"""Retrieves some metadata"""
		return self._values[key]

	def __getitem__(self, key):
		"""Shortcut to get_value"""
		return self._values[key]

	def __setitem__(self, key, value):
		"""Shortcut to set_value"""
		self.set_value(key, value)

	@staticmethod
	def build_button(label, icon_name=None, icon_widget=None, use_stock=False):
		"""Builds button situable for action area"""
		b = Gtk.Button.new_from_stock(label) if use_stock else Gtk.Button.new_with_label(label)
		b.set_use_underline(True)
		if icon_name is not None:
			icon_widget = Gtk.Image()
			icon_widget.set_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
		if icon_widget is not None:
			b.set_image(icon_widget)
			b.set_always_show_image(True)
		return b



"""SC Controller - ActionParser

Parses action(s) expressed as string or in dict loaded from json file into
one or more Action instances.
"""

from __future__ import annotations

import sys
import token as TokenType
from tokenize import TokenError, generate_tokens
from typing import NamedTuple, Self

from scc.actions import Action, MultiAction, NoAction, RangeOP
from scc.constants import PARSER_CONSTANTS, STICK, HapticPos, SCButtons
from scc.macros import Macro
from scc.tools import nameof
from scc.uinput import Axes, Keys, Rels


class ParseError(Exception):
	pass


def build_action_constants() -> dict:
	"""Generate dicts for ActionParser.CONSTS."""
	rv = {
		"Keys": Keys,
		"Axes": Axes,
		"Rels": Rels,
		"HapticPos": HapticPos,
		"None": NoAction(),
		"True": True,
		"False": False,
	}
	for c in PARSER_CONSTANTS:
		rv[c] = c
	for tpl in (Keys, Axes, Rels, SCButtons, HapticPos):
		for x in tpl:
			rv[x.name] = x
	for b in ("A", "B", "X", "Y", "START", "SELECT"):
		name = f"BTN_{b}"
		rv[name] = getattr(Keys, name)
	return rv


class ActionParser:
	"""Parse action expressed as string into Action instances.

	Usage:
		ap = ActionParser(string)
		action = ap.parse()
		if action is None:
			error = ap.get_error()
			# do something with error
	"""

	class Token(NamedTuple):
		type: int
		value: str

	CONSTS = build_action_constants()

	def __init__(self, string: str = "") -> None:
		self.restart(string)
		self.tokens: list[ActionParser.Token] | None

	def from_json_data(self, data: dict, key: str | None = None):
		"""Convert dict stored in profile file into action.

		May throw ParseError.
		"""
		if key is not None:
			# Don't fail if called for non-existent key, return NoAction instead.
			# Using this is shorter than calling 'if button in data["buttons"]: ...' everywhere
			if key in data:
				return self.from_json_data(data[key], None)
			return NoAction()

		a = self.restart(data["action"]).parse() or NoAction() if "action" in data else NoAction()
		decoders = set()
		for data_key in data:
			if data_key in Action.PKEYS:
				decoders.add(Action.PKEYS[data_key])

		if decoders:
			for cls in sorted(decoders, key=lambda a: a.PROFILE_KEY_PRIORITY):
				a = cls.decode(data, a, self, 0)  # Profile version is not yet used anywhere
		return a

	def restart(self, s: str | bytes) -> Self:
		"""Restart parsing with a new string.

		Returns self for chaining.
		"""
		if type(s) is bytes:
			s = s.decode("utf-8")

		try:
			self.tokens = [
				ActionParser.Token(token_type, string)
				for token_type, string, *_ in generate_tokens(iter([s]).__next__)
				if token_type != TokenType.ENDMARKER
			]
		except TokenError:
			self.tokens = None
		self.index = 0
		return self

	def _next_token(self) -> Token:
		if self.tokens is None:
			sys.exit("This shouldn't happen, self.tokens is none")
		rv = self.tokens[self.index]
		self.index += 1
		return rv

	def _peek_token(self) -> Token:
		"""As _next_token, but without increasing counter."""
		if self.tokens is None:
			sys.exit("This shouldn't happen, self.tokens is none")
		return self.tokens[self.index]

	def _tokens_left(self) -> bool:
		"""Return True if there are any tokens left."""
		if self.tokens is None:
			sys.exit("This shouldn't happen, self.tokens is none")
		return self.index < len(self.tokens)

	def _parse_parameter(self):
		"""Parse a single parameter."""
		t = self._next_token()
		while t.type in (TokenType.NL, TokenType.NEWLINE) or t.value == "\n":
			if not self._tokens_left():
				raise ParseError("Expected parameter at end of string")
			t = self._next_token()

		if t.type == TokenType.NAME:
			# Constant or action used as parameter
			if self._tokens_left() and self._peek_token().type == TokenType.OP and self._peek_token().value == "(":
				# Action used as parameter
				self.index -= 1  # go step back and reparse as action
				parameter = self._parse_action()
			elif (
				self._tokens_left()
				and t.value in Action.ALL
				and type(Action.ALL[t.value]) is dict
				and self._peek_token().value == "."
			):
				# SOMETHING.Action used as parameter
				self.index -= 1  # go step back and reparse as action
				parameter = self._parse_action()
			else:
				# Constant
				if t.value not in ActionParser.CONSTS:
					raise ParseError("Expected parameter, got '%s' which is not defined" % (t.value,))
				parameter = ActionParser.CONSTS[t.value]

			# Check for dots
			while self._tokens_left() and self._peek_token().type == TokenType.OP and self._peek_token().value == ".":
				self._next_token()
				if not self._tokens_left():
					raise ParseError("Expected NAME after '.'")

				t = self._next_token()
				if not hasattr(parameter, t.value):
					raise ParseError(
						"%s has no attribute '%s'"
						% (
							parameter,
							t.value,
						),
					)
				parameter = getattr(parameter, t.value)

			# Check for ranges (<, >, <=, >=)
			if self._tokens_left() and self._peek_token().type == TokenType.OP:
				if self._peek_token().value in RangeOP.OPS:
					op = self._next_token().value
					# TODO: Maybe other axes
					if parameter not in (STICK, SCButtons.LT, SCButtons.RT, SCButtons.X, SCButtons.Y):
						raise ParseError("'%s' is not trigger nor axis" % (nameof(parameter),))
					if not self._tokens_left():
						raise ParseError("Excepted number after '%s'" % (op,))
					try:
						number = float(self._next_token().value)
					except ValueError:
						raise ParseError("Excepted number after '%s'" % (op,))
					parameter = RangeOP(parameter, op, number)

			return parameter

		if t.type == TokenType.OP and t.value == "-":
			if not self._tokens_left() or self._peek_token().type != TokenType.NUMBER:
				raise ParseError("Expected number after '-'")
			return -self._parse_number()

		if t.type == TokenType.NUMBER:
			self.index -= 1
			return self._parse_number()

		if t.type == TokenType.STRING:
			return t.value[1:-1]

		raise ParseError("Expected parameter, got '%s'" % (t.value,))

	def _parse_number(self) -> float | int:
		t = self._next_token()
		if t.type != TokenType.NUMBER:
			raise ParseError("Expected number, got '%s'" % (t.value,))
		if "." in t.value or "e" in t.value.lower():
			return float(t.value)
		if t.value.lower().startswith("0x"):
			return int(t.value, 16)
		if t.value.lower().startswith("0b"):
			return int(t.value, 2)
		return int(t.value)

	def _parse_parameters(self) -> list:
		"""Parse a parameter list."""
		# Check and skip over '('
		t = self._next_token()
		if t.type != TokenType.OP or t.value != "(":
			raise ParseError("Expected '(' of parameter list, got '%s'" % (t.value,))

		parameters = []
		while self._tokens_left():
			# Check for ')' that would end parameter list
			t = self._peek_token()
			if t.type == TokenType.OP and t.value == ")":
				self._next_token()
				return parameters

			# Parse one parameter
			parameters.append(self._parse_parameter())
			# Check if next token is either ')' or ','
			t = self._peek_token()
			while t.type in (TokenType.NL, TokenType.NEWLINE) or t.value == "\n":
				self._next_token()
				if not self._tokens_left():
					raise ParseError("Expected ',' or end of parameter list after parameter '%s'" % (parameters[-1],))
				t = self._peek_token()
			if t.type == TokenType.OP and t.value == ")":
				pass
			elif t.type == TokenType.OP and t.value == ",":
				self._next_token()
			else:
				raise ParseError("Expected ',' or end of parameter list after parameter '%s'" % (parameters[-1],))

		# Code shouldn't reach here, unless there is not closing ')' in parameter list
		raise ParseError("Unmatched parenthesis")

	def _create_action(self, cls, *pars):
		try:
			return cls(*pars)
		except ValueError as e:
			raise ParseError(str(e))
		except TypeError as e:
			print(e, file=sys.stderr)
			raise ParseError("Invalid number of parameters for '%s'" % (cls.COMMAND))

	def _parse_action(self, frm: dict = Action.ALL):
		"""Parse one action, that is one of: something(params) | something() | something."""
		# Check if next token is TokenType.NAME and grab action name from it
		t = self._next_token()
		if t.type != TokenType.NAME:
			raise ParseError("Expected action name, got '%s'" % (t.value,))
		if t.value not in frm:
			raise ParseError("Unknown action '%s'" % (t.value,))
		action_name = t.value
		action_class = frm[action_name]
		# Check if there are any tokens left - if not, return action without parameters
		if not self._tokens_left():
			return self._create_action(action_class)

		# Check if token after action name is parenthesis and if yes, parse parameters from it
		t = self._peek_token()
		parameters = []
		if t.type == TokenType.OP and t.value == ".":
			# ACTION dict can have nested dicts; SOMETHING.action
			if type(action_class) is dict:
				self._next_token()
				return self._parse_action(action_class)
			raise ParseError("Unexpected '.' after '%s'" % (action_name,))
		if t.type == TokenType.OP and t.value == "(":
			parameters = self._parse_parameters()
			if not self._tokens_left():
				return self._create_action(action_class, *parameters)
			t = self._peek_token()

		# ... or, if it is one of ';', 'and' or 'or' and if yes, parse next action
		if t.type == TokenType.NAME and t.value == "and":
			# Two (or more) actions joined by 'and'
			self._next_token()
			if not self._tokens_left():
				raise ParseError("Expected action after 'and'")
			action1 = self._create_action(action_class, *parameters)
			action2 = self._parse_action()
			return MultiAction(action1, action2)

		if t.type in (TokenType.NL, TokenType.NEWLINE) or t.value == "\n":
			# Newline can be used to join actions instead of 'and'
			self._next_token()
			if not self._tokens_left():
				# Newline at end of string is not error
				return self._create_action(action_class, *parameters)
			t = self._peek_token()
			if t.type == TokenType.OP and t.value in (")", ","):
				# ')' starts next line
				return self._create_action(action_class, *parameters)
			action1 = self._create_action(action_class, *parameters)
			action2 = self._parse_action()
			return MultiAction(action1, action2)

		if t.type == TokenType.OP and t.value == ";":
			# Two (or more) actions joined by ';'
			self._next_token()
			while self._tokens_left() and self._peek_token().type in (TokenType.NL, TokenType.NEWLINE):
				self._next_token()
			if not self._tokens_left():
				# Having ';' at end of string is not actually error
				return self._create_action(action_class, *parameters)
			action1 = self._create_action(action_class, *parameters)
			action2 = self._parse_action()
			return Macro(action1, action2)

		return self._create_action(action_class, *parameters)

	def parse(self):
		"""Return parsed action.

		Throws ParseError if action cannot be parsed.
		"""
		if self.tokens is None:
			raise ParseError("Syntax error")
		a = self._parse_action()
		if self._tokens_left():
			raise ParseError("Unexpected '%s'" % (self._next_token().value,))
		return a


class TalkingActionParser(ActionParser):
	"""ActionParser that returns None when parsing fails instead of trowing exception and outputs message to stderr."""

	def restart(self, string: str):
		self.string = string
		return ActionParser.restart(self, string)

	def parse(self):
		"""Return parsed action or None if action cannot be parsed."""
		try:
			return ActionParser.parse(self)
		except ParseError as e:
			print(f"Warning: Failed to parse '{self.string}':", e, file=sys.stderr)
